from ...tl.mtproto_request import MTProtoRequest


class PhoneCallDiscarded(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    phoneCallDiscarded#50ca4de1 flags:# need_rating:flags.2?true need_debug:flags.3?true id:long reason:flags.0?PhoneCallDiscardReason duration:flags.1?int = PhoneCall"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x50ca4de1

    def __init__(self, id, need_rating=None, need_debug=None, reason=None, duration=None):
        """
        :param need_rating: Telegram type: «true».
        :param need_debug: Telegram type: «true».
        :param id: Telegram type: «long».
        :param reason: Telegram type: «PhoneCallDiscardReason».
        :param duration: Telegram type: «int».

        Constructor for PhoneCall: Instance of either PhoneCallEmpty, PhoneCallWaiting, PhoneCallRequested, PhoneCallAccepted, PhoneCall, PhoneCallDiscarded.
        """
        super().__init__()

        self.need_rating = need_rating
        self.need_debug = need_debug
        self.id = id
        self.reason = reason
        self.duration = duration

    def on_send(self, writer):
        writer.write_int(PhoneCallDiscarded.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 2) if self.need_rating else 0
        flags |= (1 << 3) if self.need_debug else 0
        flags |= (1 << 0) if self.reason else 0
        flags |= (1 << 1) if self.duration else 0
        writer.write_int(flags)

        writer.write_long(self.id)
        if self.reason:
            self.reason.on_send(writer)

        if self.duration:
            writer.write_int(self.duration)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PhoneCallDiscarded(None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 2)) != 0:
            self.need_rating = True  # Arbitrary not-None value, no need to read since it is a flag

        if (flags & (1 << 3)) != 0:
            self.need_debug = True  # Arbitrary not-None value, no need to read since it is a flag

        self.id = reader.read_long()
        if (flags & (1 << 0)) != 0:
            self.reason = reader.tgread_object()

        if (flags & (1 << 1)) != 0:
            self.duration = reader.read_int()

    def __repr__(self):
        return 'phoneCallDiscarded#50ca4de1 flags:# need_rating:flags.2?true need_debug:flags.3?true id:long reason:flags.0?PhoneCallDiscardReason duration:flags.1?int = PhoneCall'

    def __str__(self):
        return '(phoneCallDiscarded (ID: 0x50ca4de1) = (need_rating={}, need_debug={}, id={}, reason={}, duration={}))'.format(str(self.need_rating), str(self.need_debug), str(self.id), str(self.reason), str(self.duration))



from ....tl.mtproto_request import MTProtoRequest


class GetGameHighScoresRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.getGameHighScores#e822649d peer:InputPeer id:int user_id:InputUser = messages.HighScores"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xe822649d

    def __init__(self, peer, id, user_id):
        """
        :param peer: Telegram type: «InputPeer».
        :param id: Telegram type: «int».
        :param user_id: Telegram type: «InputUser».

        :returns messages.HighScores: Instance of HighScores.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.peer = peer
        self.id = id
        self.user_id = user_id

    def on_send(self, writer):
        writer.write_int(GetGameHighScoresRequest.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.write_int(self.id)
        self.user_id.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetGameHighScoresRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'messages.getGameHighScores#e822649d peer:InputPeer id:int user_id:InputUser = messages.HighScores'

    def __str__(self):
        return '(messages.getGameHighScores (ID: 0xe822649d) = (peer={}, id={}, user_id={}))'.format(str(self.peer), str(self.id), str(self.user_id))



from ...tl.mtproto_request import MTProtoRequest


class SetClientDHParamsRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    set_client_DH_params#f5045f1f nonce:int128 server_nonce:int128 encrypted_data:string = Set_client_DH_params_answer"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xf5045f1f

    def __init__(self, nonce, server_nonce, encrypted_data):
        """
        :param nonce: Telegram type: «int128».
        :param server_nonce: Telegram type: «int128».
        :param encrypted_data: Telegram type: «string».

        :returns Set_client_DH_params_answer: Instance of either DhGenOk, DhGenRetry, DhGenFail.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.nonce = nonce
        self.server_nonce = server_nonce
        self.encrypted_data = encrypted_data

    def on_send(self, writer):
        writer.write_int(SetClientDHParamsRequest.constructor_id, signed=False)
        writer.write_large_int(self.nonce, bits=128)
        writer.write_large_int(self.server_nonce, bits=128)
        writer.tgwrite_string(self.encrypted_data)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return SetClientDHParamsRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'set_client_DH_params#f5045f1f nonce:int128 server_nonce:int128 encrypted_data:string = Set_client_DH_params_answer'

    def __str__(self):
        return '(set_client_DH_params (ID: 0xf5045f1f) = (nonce={}, server_nonce={}, encrypted_data={}))'.format(str(self.nonce), str(self.server_nonce), str(self.encrypted_data))



from .check_phone import CheckPhoneRequest
from .send_code import SendCodeRequest
from .sign_up import SignUpRequest
from .sign_in import SignInRequest
from .log_out import LogOutRequest
from .reset_authorizations import ResetAuthorizationsRequest
from .send_invites import SendInvitesRequest
from .export_authorization import ExportAuthorizationRequest
from .import_authorization import ImportAuthorizationRequest
from .bind_temp_auth_key import BindTempAuthKeyRequest
from .import_bot_authorization import ImportBotAuthorizationRequest
from .check_password import CheckPasswordRequest
from .request_password_recovery import RequestPasswordRecoveryRequest
from .recover_password import RecoverPasswordRequest
from .resend_code import ResendCodeRequest
from .cancel_code import CancelCodeRequest
from .drop_temp_auth_keys import DropTempAuthKeysRequest



import gzip
from datetime import timedelta
from threading import RLock

from .. import helpers as utils
from ..crypto import AES
from ..errors import (BadMessageError, FloodWaitError,
                      RPCError, InvalidDCError)
from ..tl.all_tlobjects import tlobjects
from ..tl.types import MsgsAck
from ..utils import BinaryReader, BinaryWriter

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())


class MtProtoSender:
    """MTProto Mobile Protocol sender (https://core.telegram.org/mtproto/description)"""

    def __init__(self, transport, session):
        self._transport = transport
        self.session = session
        self._logger = logging.getLogger(__name__)

        self._need_confirmation = []  # Message IDs that need confirmation

        # Store an RLock instance to make this class safely multi-threaded
        self._lock = RLock()

        # Used when logging out, the only request that seems to use 'ack' requests
        # TODO There might be a better way to handle msgs_ack requests
        self.logging_out = False

    def connect(self):
        """Connects to the server"""
        self._transport.connect()

    def disconnect(self):
        """Disconnects from the server"""
        self._transport.close()

    # region Send and receive

    def send(self, request):
        """Sends the specified MTProtoRequest, previously sending any message
           which needed confirmation."""

        # Now only us can be using this method
        with self._lock:
            self._logger.debug('send() acquired the lock')

            # If any message needs confirmation send an AckRequest first
            if self._need_confirmation:
                msgs_ack = MsgsAck(self._need_confirmation)
                with BinaryWriter() as writer:
                    msgs_ack.on_send(writer)
                    self._send_packet(writer.get_bytes(), msgs_ack)

                del self._need_confirmation[:]

            # Finally send our packed request
            with BinaryWriter() as writer:
                request.on_send(writer)
                self._send_packet(writer.get_bytes(), request)

            # And update the saved session
            self.session.save()

        self._logger.debug('send() released the lock')

    def receive(self, request=None, timeout=timedelta(seconds=5), updates=None):
        """Receives the specified MTProtoRequest ("fills in it"
           the received data). This also restores the updates thread.
           An optional timeout can be specified to cancel the operation
           if no data has been read after its time delta.

           If 'request' is None, a single item will be read into
           the 'updates' list (which cannot be None).

           If 'request' is not None, any update received before
           reading the request's result will be put there unless
           it's None, in which case updates will be ignored.
        """
        if request is None and updates is None:
            raise ValueError('Both the "request" and "updates"'
                             'parameters cannot be None at the same time.')

        with self._lock:
            self._logger.debug('receive() acquired the lock')
            # Don't stop trying to receive until we get the request we wanted
            # or, if there is no request, until we read an update
            while True:
                self._logger.info('Trying to .receive() the request result...')
                seq, body = self._transport.receive(timeout)
                message, remote_msg_id, remote_sequence = self._decode_msg(body)

                with BinaryReader(message) as reader:
                    self._process_msg(remote_msg_id, remote_sequence, reader,
                                      request, updates)

                if request is None:
                    if updates:
                        break  # No request but one update read, exit
                elif request.confirm_received:
                        break  # Request, and result read, exit

            self._logger.info('Request result received')
        self._logger.debug('receive() released the lock')

    def receive_update(self, timeout=timedelta(seconds=5)):
        """Receives an update object and returns its result"""
        updates = []
        self.receive(timeout=timeout, updates=updates)
        return updates[0]

    def cancel_receive(self):
        """Cancels any pending receive operation
           by raising a ReadCancelledError"""
        self._transport.cancel_receive()

    # endregion

    # region Low level processing

    def _send_packet(self, packet, request):
        """Sends the given packet bytes with the additional
           information of the original request. This does NOT lock the threads!"""
        request.msg_id = self.session.get_new_msg_id()

        # First calculate plain_text to encrypt it
        with BinaryWriter() as plain_writer:
            plain_writer.write_long(self.session.salt, signed=False)
            plain_writer.write_long(self.session.id, signed=False)
            plain_writer.write_long(request.msg_id)
            plain_writer.write_int(
                self.session.generate_sequence(request.confirmed))

            plain_writer.write_int(len(packet))
            plain_writer.write(packet)

            msg_key = utils.calc_msg_key(plain_writer.get_bytes())

            key, iv = utils.calc_key(self.session.auth_key.key, msg_key, True)
            cipher_text = AES.encrypt_ige(plain_writer.get_bytes(), key, iv)

        # And then finally send the encrypted packet
        with BinaryWriter() as cipher_writer:
            cipher_writer.write_long(
                self.session.auth_key.key_id, signed=False)
            cipher_writer.write(msg_key)
            cipher_writer.write(cipher_text)
            self._transport.send(cipher_writer.get_bytes())

    def _decode_msg(self, body):
        """Decodes an received encrypted message body bytes"""
        message = None
        remote_msg_id = None
        remote_sequence = None

        with BinaryReader(body) as reader:
            if len(body) < 8:
                raise BufferError("Can't decode packet ({})".format(body))

            # TODO Check for both auth key ID and msg_key correctness
            reader.read_long()  # remote_auth_key_id
            msg_key = reader.read(16)

            key, iv = utils.calc_key(self.session.auth_key.key, msg_key, False)
            plain_text = AES.decrypt_ige(
                reader.read(len(body) - reader.tell_position()), key, iv)

            with BinaryReader(plain_text) as plain_text_reader:
                plain_text_reader.read_long()  # remote_salt
                plain_text_reader.read_long()  # remote_session_id
                remote_msg_id = plain_text_reader.read_long()
                remote_sequence = plain_text_reader.read_int()
                msg_len = plain_text_reader.read_int()
                message = plain_text_reader.read(msg_len)

        return message, remote_msg_id, remote_sequence

    def _process_msg(self, msg_id, sequence, reader, request, updates):
        """Processes and handles a Telegram message"""

        # TODO Check salt, session_id and sequence_number
        self._need_confirmation.append(msg_id)

        code = reader.read_int(signed=False)
        reader.seek(-4)

        # The following codes are "parsed manually"
        if code == 0xf35c6d01:  # rpc_result, (response of an RPC call, i.e., we sent a request)
            return self._handle_rpc_result(
                msg_id, sequence, reader, request)

        if code == 0x347773c5:  # pong
            return self._handle_pong(
                msg_id, sequence, reader, request)

        if code == 0x73f1f8dc:  # msg_container
            return self._handle_container(
                msg_id, sequence, reader, request, updates)

        if code == 0x3072cfa1:  # gzip_packed
            return self._handle_gzip_packed(
                msg_id, sequence, reader, request, updates)

        if code == 0xedab447b:  # bad_server_salt
            return self._handle_bad_server_salt(
                msg_id, sequence, reader, request)

        if code == 0xa7eff811:  # bad_msg_notification
            return self._handle_bad_msg_notification(
                msg_id, sequence, reader)

        # msgs_ack, it may handle the request we wanted
        if code == 0x62d6b459:
            ack = reader.tgread_object()
            if request and request.msg_id in ack.msg_ids:
                self._logger.warning('Ack found for the current request ID')

                if self.logging_out:
                    self._logger.info('Message ack confirmed the logout request')
                    request.confirm_received = True

            return False

        # If the code is not parsed manually, then it was parsed by the code generator!
        # In this case, we will simply treat the incoming TLObject as an Update,
        # if we can first find a matching TLObject
        if code in tlobjects:
            result = reader.tgread_object()
            if updates is None:
                self._logger.debug('Ignored update for %s', repr(result))
            else:
                self._logger.debug('Read update for %s', repr(result))
                updates.append(result)

            return False

        self._logger.warning('Unknown message: {}'.format(hex(code)))
        return False

    # endregion

    # region Message handling

    def _handle_pong(self, msg_id, sequence, reader, request):
        self._logger.debug('Handling pong')
        reader.read_int(signed=False)  # code
        received_msg_id = reader.read_long(signed=False)

        if received_msg_id == request.msg_id:
            self._logger.warning('Pong confirmed a request')
            request.confirm_received = True

        return False

    def _handle_container(self, msg_id, sequence, reader, request, updates):
        self._logger.debug('Handling container')
        reader.read_int(signed=False)  # code
        size = reader.read_int()
        for _ in range(size):
            inner_msg_id = reader.read_long(signed=False)
            reader.read_int()  # inner_sequence
            inner_length = reader.read_int()
            begin_position = reader.tell_position()

            # Note that this code is IMPORTANT for skipping RPC results of
            # lost requests (i.e., ones from the previous connection session)
            if not self._process_msg(
                    inner_msg_id, sequence, reader, request, updates):
                reader.set_position(begin_position + inner_length)

        return False

    def _handle_bad_server_salt(self, msg_id, sequence, reader, request):
        self._logger.debug('Handling bad server salt')
        reader.read_int(signed=False)  # code
        reader.read_long(signed=False)  # bad_msg_id
        reader.read_int()  # bad_msg_seq_no
        reader.read_int()  # error_code
        new_salt = reader.read_long(signed=False)

        self.session.salt = new_salt

        if request is None:
            raise ValueError(
                'Tried to handle a bad server salt with no request specified')

        # Resend
        self.send(request)

        return True

    def _handle_bad_msg_notification(self, msg_id, sequence, reader):
        self._logger.debug('Handling bad message notification')
        reader.read_int(signed=False)  # code
        reader.read_long(signed=False)  # request_id
        reader.read_int()  # request_sequence

        error_code = reader.read_int()
        error = BadMessageError(error_code)
        if error_code in (16, 17):
            # sent msg_id too low or too high (respectively).
            # Use the current msg_id to determine the right time offset.
            self.session.update_time_offset(correct_msg_id=msg_id)
            self.session.save()
            self._logger.warning('Read Bad Message error: ' + str(error))
            self._logger.info('Attempting to use the correct time offset.')
        else:
            raise error

    def _handle_rpc_result(self, msg_id, sequence, reader, request):
        self._logger.debug('Handling RPC result, request is%s None', ' not' if request else '')
        reader.read_int(signed=False)  # code
        request_id = reader.read_long(signed=False)
        inner_code = reader.read_int(signed=False)

        if request and request_id == request.msg_id:
            request.confirm_received = True

        if inner_code == 0x2144ca19:  # RPC Error
            error = RPCError(
                code=reader.read_int(), message=reader.tgread_string())

            self._logger.warning('Read RPC error: %s', str(error))
            if error.must_resend:
                if not request:
                    raise ValueError(
                        'The previously sent request must be resent. '
                        'However, no request was previously sent '
                        '(possibly called from a different thread).')
                request.confirm_received = False

            if error.message.startswith('FLOOD_WAIT_'):
                self._updates_thread_sleep = error.additional_data
                raise FloodWaitError(seconds=error.additional_data)

            elif '_MIGRATE_' in error.message:
                raise InvalidDCError(error)

            else:
                raise error
        else:
            if not request:
                raise ValueError(
                    'The request needed to read this RPC result was not '
                    'found (possibly called receive() from another thread).')

            self._logger.debug('Reading request response')
            if inner_code == 0x3072cfa1:  # GZip packed
                unpacked_data = gzip.decompress(reader.tgread_bytes())
                with BinaryReader(unpacked_data) as compressed_reader:
                    request.on_response(compressed_reader)
            else:
                reader.seek(-4)
                if request_id == request.msg_id:
                    request.on_response(reader)
                else:
                    # note: if it's really a result for RPC from previous connection
                    # session, it will be skipped by the handle_container()
                    self._logger.warning('RPC result found for unknown request (maybe from previous connection session)')

    def _handle_gzip_packed(self, msg_id, sequence, reader, request, updates):
        self._logger.debug('Handling gzip packed data')
        reader.read_int(signed=False)  # code
        packed_data = reader.tgread_bytes()
        unpacked_data = gzip.decompress(packed_data)

        with BinaryReader(unpacked_data) as compressed_reader:
            return self._process_msg(msg_id, sequence, compressed_reader,
                                     request, updates)

    # endregion



from ...tl.mtproto_request import MTProtoRequest


class Photo(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    photo#9288dd29 flags:# has_stickers:flags.0?true id:long access_hash:long date:date sizes:Vector<PhotoSize> = Photo"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x9288dd29

    def __init__(self, id, access_hash, date, sizes, has_stickers=None):
        """
        :param has_stickers: Telegram type: «true».
        :param id: Telegram type: «long».
        :param access_hash: Telegram type: «long».
        :param date: Telegram type: «date».
        :param sizes: Telegram type: «PhotoSize». Must be a list.

        Constructor for Photo: Instance of either PhotoEmpty, Photo.
        """
        super().__init__()

        self.has_stickers = has_stickers
        self.id = id
        self.access_hash = access_hash
        self.date = date
        self.sizes = sizes

    def on_send(self, writer):
        writer.write_int(Photo.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.has_stickers else 0
        writer.write_int(flags)

        writer.write_long(self.id)
        writer.write_long(self.access_hash)
        writer.tgwrite_date(self.date)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.sizes))
        for sizes_item in self.sizes:
            sizes_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return Photo(None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.has_stickers = True  # Arbitrary not-None value, no need to read since it is a flag

        self.id = reader.read_long()
        self.access_hash = reader.read_long()
        self.date = reader.tgread_date()
        reader.read_int()  # Vector's constructor ID
        self.sizes = []  # Initialize an empty list
        sizes_len = reader.read_int()
        for _ in range(sizes_len):
            sizes_item = reader.tgread_object()
            self.sizes.append(sizes_item)

    def __repr__(self):
        return 'photo#9288dd29 flags:# has_stickers:flags.0?true id:long access_hash:long date:date sizes:Vector<PhotoSize> = Photo'

    def __str__(self):
        return '(photo (ID: 0x9288dd29) = (has_stickers={}, id={}, access_hash={}, date={}, sizes={}))'.format(str(self.has_stickers), str(self.id), str(self.access_hash), str(self.date), None if not self.sizes else [str(_) for _ in self.sizes])



from ...tl.mtproto_request import MTProtoRequest


class InputMediaInvoice(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inputMediaInvoice#92153685 flags:# title:string description:string photo:flags.0?InputWebDocument invoice:Invoice payload:bytes provider:string start_param:string = InputMedia"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x92153685

    def __init__(self, title, description, invoice, payload, provider, start_param, photo=None):
        """
        :param title: Telegram type: «string».
        :param description: Telegram type: «string».
        :param photo: Telegram type: «InputWebDocument».
        :param invoice: Telegram type: «Invoice».
        :param payload: Telegram type: «bytes».
        :param provider: Telegram type: «string».
        :param start_param: Telegram type: «string».

        Constructor for InputMedia: Instance of either InputMediaEmpty, InputMediaUploadedPhoto, InputMediaPhoto, InputMediaGeoPoint, InputMediaContact, InputMediaUploadedDocument, InputMediaUploadedThumbDocument, InputMediaDocument, InputMediaVenue, InputMediaGifExternal, InputMediaPhotoExternal, InputMediaDocumentExternal, InputMediaGame, InputMediaInvoice.
        """
        super().__init__()

        self.title = title
        self.description = description
        self.photo = photo
        self.invoice = invoice
        self.payload = payload
        self.provider = provider
        self.start_param = start_param

    def on_send(self, writer):
        writer.write_int(InputMediaInvoice.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.photo else 0
        writer.write_int(flags)

        writer.tgwrite_string(self.title)
        writer.tgwrite_string(self.description)
        if self.photo:
            self.photo.on_send(writer)

        self.invoice.on_send(writer)
        writer.tgwrite_bytes(self.payload)
        writer.tgwrite_string(self.provider)
        writer.tgwrite_string(self.start_param)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InputMediaInvoice(None, None, None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        self.title = reader.tgread_string()
        self.description = reader.tgread_string()
        if (flags & (1 << 0)) != 0:
            self.photo = reader.tgread_object()

        self.invoice = reader.tgread_object()
        self.payload = reader.tgread_bytes()
        self.provider = reader.tgread_string()
        self.start_param = reader.tgread_string()

    def __repr__(self):
        return 'inputMediaInvoice#92153685 flags:# title:string description:string photo:flags.0?InputWebDocument invoice:Invoice payload:bytes provider:string start_param:string = InputMedia'

    def __str__(self):
        return '(inputMediaInvoice (ID: 0x92153685) = (title={}, description={}, photo={}, invoice={}, payload={}, provider={}, start_param={}))'.format(str(self.title), str(self.description), str(self.photo), str(self.invoice), str(self.payload), str(self.provider), str(self.start_param))



from ...tl.mtproto_request import MTProtoRequest


class PageBlockFooter(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    pageBlockFooter#48870999 text:RichText = PageBlock"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x48870999

    def __init__(self, text):
        """
        :param text: Telegram type: «RichText».

        Constructor for PageBlock: Instance of either PageBlockUnsupported, PageBlockTitle, PageBlockSubtitle, PageBlockAuthorDate, PageBlockHeader, PageBlockSubheader, PageBlockParagraph, PageBlockPreformatted, PageBlockFooter, PageBlockDivider, PageBlockAnchor, PageBlockList, PageBlockBlockquote, PageBlockPullquote, PageBlockPhoto, PageBlockVideo, PageBlockCover, PageBlockEmbed, PageBlockEmbedPost, PageBlockCollage, PageBlockSlideshow, PageBlockChannel.
        """
        super().__init__()

        self.text = text

    def on_send(self, writer):
        writer.write_int(PageBlockFooter.constructor_id, signed=False)
        self.text.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PageBlockFooter(None)

    def on_response(self, reader):
        self.text = reader.tgread_object()

    def __repr__(self):
        return 'pageBlockFooter#48870999 text:RichText = PageBlock'

    def __str__(self):
        return '(pageBlockFooter (ID: 0x48870999) = (text={}))'.format(str(self.text))



from ....tl.mtproto_request import MTProtoRequest


class SendCustomRequestRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    bots.sendCustomRequest#aa2769ed custom_method:string params:DataJSON = DataJSON"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xaa2769ed

    def __init__(self, custom_method, params):
        """
        :param custom_method: Telegram type: «string».
        :param params: Telegram type: «DataJSON».

        :returns DataJSON: Instance of DataJSON.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.custom_method = custom_method
        self.params = params

    def on_send(self, writer):
        writer.write_int(SendCustomRequestRequest.constructor_id, signed=False)
        writer.tgwrite_string(self.custom_method)
        self.params.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return SendCustomRequestRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'bots.sendCustomRequest#aa2769ed custom_method:string params:DataJSON = DataJSON'

    def __str__(self):
        return '(bots.sendCustomRequest (ID: 0xaa2769ed) = (custom_method={}, params={}))'.format(str(self.custom_method), str(self.params))



from ...tl.mtproto_request import MTProtoRequest


class MessageActionPhoneCall(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messageActionPhoneCall#80e11a7f flags:# call_id:long reason:flags.0?PhoneCallDiscardReason duration:flags.1?int = MessageAction"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x80e11a7f

    def __init__(self, call_id, reason=None, duration=None):
        """
        :param call_id: Telegram type: «long».
        :param reason: Telegram type: «PhoneCallDiscardReason».
        :param duration: Telegram type: «int».

        Constructor for MessageAction: Instance of either MessageActionEmpty, MessageActionChatCreate, MessageActionChatEditTitle, MessageActionChatEditPhoto, MessageActionChatDeletePhoto, MessageActionChatAddUser, MessageActionChatDeleteUser, MessageActionChatJoinedByLink, MessageActionChannelCreate, MessageActionChatMigrateTo, MessageActionChannelMigrateFrom, MessageActionPinMessage, MessageActionHistoryClear, MessageActionGameScore, MessageActionPaymentSentMe, MessageActionPaymentSent, MessageActionPhoneCall.
        """
        super().__init__()

        self.call_id = call_id
        self.reason = reason
        self.duration = duration

    def on_send(self, writer):
        writer.write_int(MessageActionPhoneCall.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.reason else 0
        flags |= (1 << 1) if self.duration else 0
        writer.write_int(flags)

        writer.write_long(self.call_id)
        if self.reason:
            self.reason.on_send(writer)

        if self.duration:
            writer.write_int(self.duration)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return MessageActionPhoneCall(None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        self.call_id = reader.read_long()
        if (flags & (1 << 0)) != 0:
            self.reason = reader.tgread_object()

        if (flags & (1 << 1)) != 0:
            self.duration = reader.read_int()

    def __repr__(self):
        return 'messageActionPhoneCall#80e11a7f flags:# call_id:long reason:flags.0?PhoneCallDiscardReason duration:flags.1?int = MessageAction'

    def __str__(self):
        return '(messageActionPhoneCall (ID: 0x80e11a7f) = (call_id={}, reason={}, duration={}))'.format(str(self.call_id), str(self.reason), str(self.duration))



from ...tl.mtproto_request import MTProtoRequest


class BotCommand(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    botCommand#c27ac8c7 command:string description:string = BotCommand"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xc27ac8c7

    def __init__(self, command, description):
        """
        :param command: Telegram type: «string».
        :param description: Telegram type: «string».

        Constructor for BotCommand: Instance of BotCommand.
        """
        super().__init__()

        self.command = command
        self.description = description

    def on_send(self, writer):
        writer.write_int(BotCommand.constructor_id, signed=False)
        writer.tgwrite_string(self.command)
        writer.tgwrite_string(self.description)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return BotCommand(None, None)

    def on_response(self, reader):
        self.command = reader.tgread_string()
        self.description = reader.tgread_string()

    def __repr__(self):
        return 'botCommand#c27ac8c7 command:string description:string = BotCommand'

    def __str__(self):
        return '(botCommand (ID: 0xc27ac8c7) = (command={}, description={}))'.format(str(self.command), str(self.description))



from ...tl.mtproto_request import MTProtoRequest


class InputMediaDocumentExternal(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inputMediaDocumentExternal#e5e9607c url:string caption:string = InputMedia"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xe5e9607c

    def __init__(self, url, caption):
        """
        :param url: Telegram type: «string».
        :param caption: Telegram type: «string».

        Constructor for InputMedia: Instance of either InputMediaEmpty, InputMediaUploadedPhoto, InputMediaPhoto, InputMediaGeoPoint, InputMediaContact, InputMediaUploadedDocument, InputMediaUploadedThumbDocument, InputMediaDocument, InputMediaVenue, InputMediaGifExternal, InputMediaPhotoExternal, InputMediaDocumentExternal, InputMediaGame, InputMediaInvoice.
        """
        super().__init__()

        self.url = url
        self.caption = caption

    def on_send(self, writer):
        writer.write_int(InputMediaDocumentExternal.constructor_id, signed=False)
        writer.tgwrite_string(self.url)
        writer.tgwrite_string(self.caption)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InputMediaDocumentExternal(None, None)

    def on_response(self, reader):
        self.url = reader.tgread_string()
        self.caption = reader.tgread_string()

    def __repr__(self):
        return 'inputMediaDocumentExternal#e5e9607c url:string caption:string = InputMedia'

    def __str__(self):
        return '(inputMediaDocumentExternal (ID: 0xe5e9607c) = (url={}, caption={}))'.format(str(self.url), str(self.caption))



from ....tl.mtproto_request import MTProtoRequest


class SavedGifsNotModified(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.savedGifsNotModified#e8025ca2  = messages.SavedGifs"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xe8025ca2

    def __init__(self):
        super().__init__()

    def on_send(self, writer):
        writer.write_int(SavedGifsNotModified.constructor_id, signed=False)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return SavedGifsNotModified()

    def on_response(self, reader):
        pass

    def __repr__(self):
        return 'messages.savedGifsNotModified#e8025ca2  = messages.SavedGifs'

    def __str__(self):
        return '(messages.savedGifsNotModified (ID: 0xe8025ca2) = ())'.format()



from ...tl.mtproto_request import MTProtoRequest


class UpdatePhoneCall(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updatePhoneCall#ab0f6b1e phone_call:PhoneCall = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xab0f6b1e

    def __init__(self, phone_call):
        """
        :param phone_call: Telegram type: «PhoneCall».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.phone_call = phone_call

    def on_send(self, writer):
        writer.write_int(UpdatePhoneCall.constructor_id, signed=False)
        self.phone_call.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdatePhoneCall(None)

    def on_response(self, reader):
        self.phone_call = reader.tgread_object()

    def __repr__(self):
        return 'updatePhoneCall#ab0f6b1e phone_call:PhoneCall = Update'

    def __str__(self):
        return '(updatePhoneCall (ID: 0xab0f6b1e) = (phone_call={}))'.format(str(self.phone_call))



from ...tl.mtproto_request import MTProtoRequest


class UpdateNotifySettings(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateNotifySettings#bec268ef peer:NotifyPeer notify_settings:PeerNotifySettings = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xbec268ef

    def __init__(self, peer, notify_settings):
        """
        :param peer: Telegram type: «NotifyPeer».
        :param notify_settings: Telegram type: «PeerNotifySettings».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.peer = peer
        self.notify_settings = notify_settings

    def on_send(self, writer):
        writer.write_int(UpdateNotifySettings.constructor_id, signed=False)
        self.peer.on_send(writer)
        self.notify_settings.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateNotifySettings(None, None)

    def on_response(self, reader):
        self.peer = reader.tgread_object()
        self.notify_settings = reader.tgread_object()

    def __repr__(self):
        return 'updateNotifySettings#bec268ef peer:NotifyPeer notify_settings:PeerNotifySettings = Update'

    def __str__(self):
        return '(updateNotifySettings (ID: 0xbec268ef) = (peer={}, notify_settings={}))'.format(str(self.peer), str(self.notify_settings))



from ....tl.mtproto_request import MTProtoRequest


class ChatsSlice(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.chatsSlice#9cd81144 count:int chats:Vector<Chat> = messages.Chats"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x9cd81144

    def __init__(self, count, chats):
        """
        :param count: Telegram type: «int».
        :param chats: Telegram type: «Chat». Must be a list.

        Constructor for messages.Chats: Instance of either Chats, ChatsSlice.
        """
        super().__init__()

        self.count = count
        self.chats = chats

    def on_send(self, writer):
        writer.write_int(ChatsSlice.constructor_id, signed=False)
        writer.write_int(self.count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for chats_item in self.chats:
            chats_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return ChatsSlice(None, None)

    def on_response(self, reader):
        self.count = reader.read_int()
        reader.read_int()  # Vector's constructor ID
        self.chats = []  # Initialize an empty list
        chats_len = reader.read_int()
        for _ in range(chats_len):
            chats_item = reader.tgread_object()
            self.chats.append(chats_item)

    def __repr__(self):
        return 'messages.chatsSlice#9cd81144 count:int chats:Vector<Chat> = messages.Chats'

    def __str__(self):
        return '(messages.chatsSlice (ID: 0x9cd81144) = (count={}, chats={}))'.format(str(self.count), None if not self.chats else [str(_) for _ in self.chats])



from ...tl.mtproto_request import MTProtoRequest


class PageBlockCover(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    pageBlockCover#39f23300 cover:PageBlock = PageBlock"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x39f23300

    def __init__(self, cover):
        """
        :param cover: Telegram type: «PageBlock».

        Constructor for PageBlock: Instance of either PageBlockUnsupported, PageBlockTitle, PageBlockSubtitle, PageBlockAuthorDate, PageBlockHeader, PageBlockSubheader, PageBlockParagraph, PageBlockPreformatted, PageBlockFooter, PageBlockDivider, PageBlockAnchor, PageBlockList, PageBlockBlockquote, PageBlockPullquote, PageBlockPhoto, PageBlockVideo, PageBlockCover, PageBlockEmbed, PageBlockEmbedPost, PageBlockCollage, PageBlockSlideshow, PageBlockChannel.
        """
        super().__init__()

        self.cover = cover

    def on_send(self, writer):
        writer.write_int(PageBlockCover.constructor_id, signed=False)
        self.cover.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PageBlockCover(None)

    def on_response(self, reader):
        self.cover = reader.tgread_object()

    def __repr__(self):
        return 'pageBlockCover#39f23300 cover:PageBlock = PageBlock'

    def __str__(self):
        return '(pageBlockCover (ID: 0x39f23300) = (cover={}))'.format(str(self.cover))



from ...tl.mtproto_request import MTProtoRequest


class TextConcat(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    textConcat#7e6260d7 texts:Vector<RichText> = RichText"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x7e6260d7

    def __init__(self, texts):
        """
        :param texts: Telegram type: «RichText». Must be a list.

        Constructor for RichText: Instance of either TextEmpty, TextPlain, TextBold, TextItalic, TextUnderline, TextStrike, TextFixed, TextUrl, TextEmail, TextConcat.
        """
        super().__init__()

        self.texts = texts

    def on_send(self, writer):
        writer.write_int(TextConcat.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.texts))
        for texts_item in self.texts:
            texts_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return TextConcat(None)

    def on_response(self, reader):
        reader.read_int()  # Vector's constructor ID
        self.texts = []  # Initialize an empty list
        texts_len = reader.read_int()
        for _ in range(texts_len):
            texts_item = reader.tgread_object()
            self.texts.append(texts_item)

    def __repr__(self):
        return 'textConcat#7e6260d7 texts:Vector<RichText> = RichText'

    def __str__(self):
        return '(textConcat (ID: 0x7e6260d7) = (texts={}))'.format(None if not self.texts else [str(_) for _ in self.texts])



from ...tl.mtproto_request import MTProtoRequest


class InputBotInlineMessageID(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inputBotInlineMessageID#890c3d89 dc_id:int id:long access_hash:long = InputBotInlineMessageID"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x890c3d89

    def __init__(self, dc_id, id, access_hash):
        """
        :param dc_id: Telegram type: «int».
        :param id: Telegram type: «long».
        :param access_hash: Telegram type: «long».

        Constructor for InputBotInlineMessageID: Instance of InputBotInlineMessageID.
        """
        super().__init__()

        self.dc_id = dc_id
        self.id = id
        self.access_hash = access_hash

    def on_send(self, writer):
        writer.write_int(InputBotInlineMessageID.constructor_id, signed=False)
        writer.write_int(self.dc_id)
        writer.write_long(self.id)
        writer.write_long(self.access_hash)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InputBotInlineMessageID(None, None, None)

    def on_response(self, reader):
        self.dc_id = reader.read_int()
        self.id = reader.read_long()
        self.access_hash = reader.read_long()

    def __repr__(self):
        return 'inputBotInlineMessageID#890c3d89 dc_id:int id:long access_hash:long = InputBotInlineMessageID'

    def __str__(self):
        return '(inputBotInlineMessageID (ID: 0x890c3d89) = (dc_id={}, id={}, access_hash={}))'.format(str(self.dc_id), str(self.id), str(self.access_hash))



from ...tl.mtproto_request import MTProtoRequest


class DocumentAttributeVideo(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    documentAttributeVideo#0ef02ce6 flags:# round_message:flags.0?true duration:int w:int h:int = DocumentAttribute"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xef02ce6

    def __init__(self, duration, w, h, round_message=None):
        """
        :param round_message: Telegram type: «true».
        :param duration: Telegram type: «int».
        :param w: Telegram type: «int».
        :param h: Telegram type: «int».

        Constructor for DocumentAttribute: Instance of either DocumentAttributeImageSize, DocumentAttributeAnimated, DocumentAttributeSticker, DocumentAttributeVideo, DocumentAttributeAudio, DocumentAttributeFilename, DocumentAttributeHasStickers.
        """
        super().__init__()

        self.round_message = round_message
        self.duration = duration
        self.w = w
        self.h = h

    def on_send(self, writer):
        writer.write_int(DocumentAttributeVideo.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.round_message else 0
        writer.write_int(flags)

        writer.write_int(self.duration)
        writer.write_int(self.w)
        writer.write_int(self.h)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return DocumentAttributeVideo(None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.round_message = True  # Arbitrary not-None value, no need to read since it is a flag

        self.duration = reader.read_int()
        self.w = reader.read_int()
        self.h = reader.read_int()

    def __repr__(self):
        return 'documentAttributeVideo#0ef02ce6 flags:# round_message:flags.0?true duration:int w:int h:int = DocumentAttribute'

    def __str__(self):
        return '(documentAttributeVideo (ID: 0xef02ce6) = (round_message={}, duration={}, w={}, h={}))'.format(str(self.round_message), str(self.duration), str(self.w), str(self.h))



from ...tl.mtproto_request import MTProtoRequest


class RpcAnswerDropped(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    rpc_answer_dropped#a43ad8b7 msg_id:long seq_no:int bytes:int = RpcDropAnswer"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xa43ad8b7

    def __init__(self, msg_id, seq_no, bytes):
        """
        :param msg_id: Telegram type: «long».
        :param seq_no: Telegram type: «int».
        :param bytes: Telegram type: «int».

        Constructor for RpcDropAnswer: Instance of either RpcAnswerUnknown, RpcAnswerDroppedRunning, RpcAnswerDropped.
        """
        super().__init__()

        self.msg_id = msg_id
        self.seq_no = seq_no
        self.bytes = bytes

    def on_send(self, writer):
        writer.write_int(RpcAnswerDropped.constructor_id, signed=False)
        writer.write_long(self.msg_id)
        writer.write_int(self.seq_no)
        writer.write_int(self.bytes)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return RpcAnswerDropped(None, None, None)

    def on_response(self, reader):
        self.msg_id = reader.read_long()
        self.seq_no = reader.read_int()
        self.bytes = reader.read_int()

    def __repr__(self):
        return 'rpc_answer_dropped#a43ad8b7 msg_id:long seq_no:int bytes:int = RpcDropAnswer'

    def __str__(self):
        return '(rpc_answer_dropped (ID: 0xa43ad8b7) = (msg_id={}, seq_no={}, bytes={}))'.format(str(self.msg_id), str(self.seq_no), str(self.bytes))



from ...tl.mtproto_request import MTProtoRequest


class PageBlockAuthorDate(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    pageBlockAuthorDate#baafe5e0 author:RichText published_date:date = PageBlock"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xbaafe5e0

    def __init__(self, author, published_date):
        """
        :param author: Telegram type: «RichText».
        :param published_date: Telegram type: «date».

        Constructor for PageBlock: Instance of either PageBlockUnsupported, PageBlockTitle, PageBlockSubtitle, PageBlockAuthorDate, PageBlockHeader, PageBlockSubheader, PageBlockParagraph, PageBlockPreformatted, PageBlockFooter, PageBlockDivider, PageBlockAnchor, PageBlockList, PageBlockBlockquote, PageBlockPullquote, PageBlockPhoto, PageBlockVideo, PageBlockCover, PageBlockEmbed, PageBlockEmbedPost, PageBlockCollage, PageBlockSlideshow, PageBlockChannel.
        """
        super().__init__()

        self.author = author
        self.published_date = published_date

    def on_send(self, writer):
        writer.write_int(PageBlockAuthorDate.constructor_id, signed=False)
        self.author.on_send(writer)
        writer.tgwrite_date(self.published_date)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PageBlockAuthorDate(None, None)

    def on_response(self, reader):
        self.author = reader.tgread_object()
        self.published_date = reader.tgread_date()

    def __repr__(self):
        return 'pageBlockAuthorDate#baafe5e0 author:RichText published_date:date = PageBlock'

    def __str__(self):
        return '(pageBlockAuthorDate (ID: 0xbaafe5e0) = (author={}, published_date={}))'.format(str(self.author), str(self.published_date))



from ....tl.mtproto_request import MTProtoRequest


class DropTempAuthKeysRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    auth.dropTempAuthKeys#8e48a188 except_auth_keys:Vector<long> = Bool"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x8e48a188

    def __init__(self, except_auth_keys):
        """
        :param except_auth_keys: Telegram type: «long». Must be a list.

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.except_auth_keys = except_auth_keys

    def on_send(self, writer):
        writer.write_int(DropTempAuthKeysRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.except_auth_keys))
        for except_auth_keys_item in self.except_auth_keys:
            writer.write_long(except_auth_keys_item)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return DropTempAuthKeysRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'auth.dropTempAuthKeys#8e48a188 except_auth_keys:Vector<long> = Bool'

    def __str__(self):
        return '(auth.dropTempAuthKeys (ID: 0x8e48a188) = (except_auth_keys={}))'.format(None if not self.except_auth_keys else [str(_) for _ in self.except_auth_keys])



from ...tl.mtproto_request import MTProtoRequest


class FutureSalt(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    future_salt#0949d9dc valid_since:int valid_until:int salt:long = FutureSalt"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x949d9dc

    def __init__(self, valid_since, valid_until, salt):
        """
        :param valid_since: Telegram type: «int».
        :param valid_until: Telegram type: «int».
        :param salt: Telegram type: «long».

        Constructor for FutureSalt: Instance of FutureSalt.
        """
        super().__init__()

        self.valid_since = valid_since
        self.valid_until = valid_until
        self.salt = salt

    def on_send(self, writer):
        writer.write_int(FutureSalt.constructor_id, signed=False)
        writer.write_int(self.valid_since)
        writer.write_int(self.valid_until)
        writer.write_long(self.salt)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return FutureSalt(None, None, None)

    def on_response(self, reader):
        self.valid_since = reader.read_int()
        self.valid_until = reader.read_int()
        self.salt = reader.read_long()

    def __repr__(self):
        return 'future_salt#0949d9dc valid_since:int valid_until:int salt:long = FutureSalt'

    def __str__(self):
        return '(future_salt (ID: 0x949d9dc) = (valid_since={}, valid_until={}, salt={}))'.format(str(self.valid_since), str(self.valid_until), str(self.salt))



from ....tl.mtproto_request import MTProtoRequest


class GetFullChannelRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    channels.getFullChannel#08736a09 channel:InputChannel = messages.ChatFull"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x8736a09

    def __init__(self, channel):
        """
        :param channel: Telegram type: «InputChannel».

        :returns messages.ChatFull: Instance of ChatFull.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.channel = channel

    def on_send(self, writer):
        writer.write_int(GetFullChannelRequest.constructor_id, signed=False)
        self.channel.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetFullChannelRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'channels.getFullChannel#08736a09 channel:InputChannel = messages.ChatFull'

    def __str__(self):
        return '(channels.getFullChannel (ID: 0x8736a09) = (channel={}))'.format(str(self.channel))



from ...tl.mtproto_request import MTProtoRequest


class UpdateEncryptedMessagesRead(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateEncryptedMessagesRead#38fe25b7 chat_id:int max_date:date date:date = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x38fe25b7

    def __init__(self, chat_id, max_date, date):
        """
        :param chat_id: Telegram type: «int».
        :param max_date: Telegram type: «date».
        :param date: Telegram type: «date».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.chat_id = chat_id
        self.max_date = max_date
        self.date = date

    def on_send(self, writer):
        writer.write_int(UpdateEncryptedMessagesRead.constructor_id, signed=False)
        writer.write_int(self.chat_id)
        writer.tgwrite_date(self.max_date)
        writer.tgwrite_date(self.date)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateEncryptedMessagesRead(None, None, None)

    def on_response(self, reader):
        self.chat_id = reader.read_int()
        self.max_date = reader.tgread_date()
        self.date = reader.tgread_date()

    def __repr__(self):
        return 'updateEncryptedMessagesRead#38fe25b7 chat_id:int max_date:date date:date = Update'

    def __str__(self):
        return '(updateEncryptedMessagesRead (ID: 0x38fe25b7) = (chat_id={}, max_date={}, date={}))'.format(str(self.chat_id), str(self.max_date), str(self.date))



from ...tl.mtproto_request import MTProtoRequest


class UpdateChannelPinnedMessage(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateChannelPinnedMessage#98592475 channel_id:int id:int = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x98592475

    def __init__(self, channel_id, id):
        """
        :param channel_id: Telegram type: «int».
        :param id: Telegram type: «int».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.channel_id = channel_id
        self.id = id

    def on_send(self, writer):
        writer.write_int(UpdateChannelPinnedMessage.constructor_id, signed=False)
        writer.write_int(self.channel_id)
        writer.write_int(self.id)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateChannelPinnedMessage(None, None)

    def on_response(self, reader):
        self.channel_id = reader.read_int()
        self.id = reader.read_int()

    def __repr__(self):
        return 'updateChannelPinnedMessage#98592475 channel_id:int id:int = Update'

    def __str__(self):
        return '(updateChannelPinnedMessage (ID: 0x98592475) = (channel_id={}, id={}))'.format(str(self.channel_id), str(self.id))



from ....tl.mtproto_request import MTProtoRequest


class ChannelDifferenceEmpty(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updates.channelDifferenceEmpty#3e11affb flags:# final:flags.0?true pts:int timeout:flags.1?int = updates.ChannelDifference"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x3e11affb

    def __init__(self, pts, final=None, timeout=None):
        """
        :param final: Telegram type: «true».
        :param pts: Telegram type: «int».
        :param timeout: Telegram type: «int».

        Constructor for updates.ChannelDifference: Instance of either ChannelDifferenceEmpty, ChannelDifferenceTooLong, ChannelDifference.
        """
        super().__init__()

        self.final = final
        self.pts = pts
        self.timeout = timeout

    def on_send(self, writer):
        writer.write_int(ChannelDifferenceEmpty.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.final else 0
        flags |= (1 << 1) if self.timeout else 0
        writer.write_int(flags)

        writer.write_int(self.pts)
        if self.timeout:
            writer.write_int(self.timeout)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return ChannelDifferenceEmpty(None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.final = True  # Arbitrary not-None value, no need to read since it is a flag

        self.pts = reader.read_int()
        if (flags & (1 << 1)) != 0:
            self.timeout = reader.read_int()

    def __repr__(self):
        return 'updates.channelDifferenceEmpty#3e11affb flags:# final:flags.0?true pts:int timeout:flags.1?int = updates.ChannelDifference'

    def __str__(self):
        return '(updates.channelDifferenceEmpty (ID: 0x3e11affb) = (final={}, pts={}, timeout={}))'.format(str(self.final), str(self.pts), str(self.timeout))



from ...tl.mtproto_request import MTProtoRequest


class MessageMediaGame(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messageMediaGame#fdb19008 game:Game = MessageMedia"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xfdb19008

    def __init__(self, game):
        """
        :param game: Telegram type: «Game».

        Constructor for MessageMedia: Instance of either MessageMediaEmpty, MessageMediaPhoto, MessageMediaGeo, MessageMediaContact, MessageMediaUnsupported, MessageMediaDocument, MessageMediaWebPage, MessageMediaVenue, MessageMediaGame, MessageMediaInvoice.
        """
        super().__init__()

        self.game = game

    def on_send(self, writer):
        writer.write_int(MessageMediaGame.constructor_id, signed=False)
        self.game.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return MessageMediaGame(None)

    def on_response(self, reader):
        self.game = reader.tgread_object()

    def __repr__(self):
        return 'messageMediaGame#fdb19008 game:Game = MessageMedia'

    def __str__(self):
        return '(messageMediaGame (ID: 0xfdb19008) = (game={}))'.format(str(self.game))



from ....tl.mtproto_request import MTProtoRequest


class ForwardMessagesRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.forwardMessages#708e0195 flags:# silent:flags.5?true background:flags.6?true with_my_score:flags.8?true from_peer:InputPeer id:Vector<int> random_id:Vector<long> to_peer:InputPeer = Updates"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x708e0195

    def __init__(self, from_peer, id, random_id, to_peer, silent=None, background=None, with_my_score=None):
        """
        :param silent: Telegram type: «true».
        :param background: Telegram type: «true».
        :param with_my_score: Telegram type: «true».
        :param from_peer: Telegram type: «InputPeer».
        :param id: Telegram type: «int». Must be a list.
        :param random_id: Telegram type: «long». Must be a list.
        :param to_peer: Telegram type: «InputPeer».

        :returns Updates: Instance of either UpdatesTooLong, UpdateShortMessage, UpdateShortChatMessage, UpdateShort, UpdatesCombined, UpdatesTg, UpdateShortSentMessage.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.silent = silent
        self.background = background
        self.with_my_score = with_my_score
        self.from_peer = from_peer
        self.id = id
        self.random_id = random_id
        self.to_peer = to_peer

    def on_send(self, writer):
        writer.write_int(ForwardMessagesRequest.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 5) if self.silent else 0
        flags |= (1 << 6) if self.background else 0
        flags |= (1 << 8) if self.with_my_score else 0
        writer.write_int(flags)

        self.from_peer.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.id))
        for id_item in self.id:
            writer.write_int(id_item)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.random_id))
        for random_id_item in self.random_id:
            writer.write_long(random_id_item)

        self.to_peer.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return ForwardMessagesRequest(None, None, None, None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'messages.forwardMessages#708e0195 flags:# silent:flags.5?true background:flags.6?true with_my_score:flags.8?true from_peer:InputPeer id:Vector<int> random_id:Vector<long> to_peer:InputPeer = Updates'

    def __str__(self):
        return '(messages.forwardMessages (ID: 0x708e0195) = (silent={}, background={}, with_my_score={}, from_peer={}, id={}, random_id={}, to_peer={}))'.format(str(self.silent), str(self.background), str(self.with_my_score), str(self.from_peer), None if not self.id else [str(_) for _ in self.id], None if not self.random_id else [str(_) for _ in self.random_id], str(self.to_peer))



from ....tl.mtproto_request import MTProtoRequest


class TmpPassword(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    account.tmpPassword#db64fd34 tmp_password:bytes valid_until:int = account.TmpPassword"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xdb64fd34

    def __init__(self, tmp_password, valid_until):
        """
        :param tmp_password: Telegram type: «bytes».
        :param valid_until: Telegram type: «int».

        Constructor for account.TmpPassword: Instance of TmpPassword.
        """
        super().__init__()

        self.tmp_password = tmp_password
        self.valid_until = valid_until

    def on_send(self, writer):
        writer.write_int(TmpPassword.constructor_id, signed=False)
        writer.tgwrite_bytes(self.tmp_password)
        writer.write_int(self.valid_until)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return TmpPassword(None, None)

    def on_response(self, reader):
        self.tmp_password = reader.tgread_bytes()
        self.valid_until = reader.read_int()

    def __repr__(self):
        return 'account.tmpPassword#db64fd34 tmp_password:bytes valid_until:int = account.TmpPassword'

    def __str__(self):
        return '(account.tmpPassword (ID: 0xdb64fd34) = (tmp_password={}, valid_until={}))'.format(str(self.tmp_password), str(self.valid_until))



from ...tl.mtproto_request import MTProtoRequest


class WallPaper(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    wallPaper#ccb03657 id:int title:string sizes:Vector<PhotoSize> color:int = WallPaper"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xccb03657

    def __init__(self, id, title, sizes, color):
        """
        :param id: Telegram type: «int».
        :param title: Telegram type: «string».
        :param sizes: Telegram type: «PhotoSize». Must be a list.
        :param color: Telegram type: «int».

        Constructor for WallPaper: Instance of either WallPaper, WallPaperSolid.
        """
        super().__init__()

        self.id = id
        self.title = title
        self.sizes = sizes
        self.color = color

    def on_send(self, writer):
        writer.write_int(WallPaper.constructor_id, signed=False)
        writer.write_int(self.id)
        writer.tgwrite_string(self.title)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.sizes))
        for sizes_item in self.sizes:
            sizes_item.on_send(writer)

        writer.write_int(self.color)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return WallPaper(None, None, None, None)

    def on_response(self, reader):
        self.id = reader.read_int()
        self.title = reader.tgread_string()
        reader.read_int()  # Vector's constructor ID
        self.sizes = []  # Initialize an empty list
        sizes_len = reader.read_int()
        for _ in range(sizes_len):
            sizes_item = reader.tgread_object()
            self.sizes.append(sizes_item)

        self.color = reader.read_int()

    def __repr__(self):
        return 'wallPaper#ccb03657 id:int title:string sizes:Vector<PhotoSize> color:int = WallPaper'

    def __str__(self):
        return '(wallPaper (ID: 0xccb03657) = (id={}, title={}, sizes={}, color={}))'.format(str(self.id), str(self.title), None if not self.sizes else [str(_) for _ in self.sizes], str(self.color))



from ...tl.mtproto_request import MTProtoRequest


class InputNotifyPeer(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inputNotifyPeer#b8bc5b0c peer:InputPeer = InputNotifyPeer"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xb8bc5b0c

    def __init__(self, peer):
        """
        :param peer: Telegram type: «InputPeer».

        Constructor for InputNotifyPeer: Instance of either InputNotifyPeer, InputNotifyUsers, InputNotifyChats, InputNotifyAll.
        """
        super().__init__()

        self.peer = peer

    def on_send(self, writer):
        writer.write_int(InputNotifyPeer.constructor_id, signed=False)
        self.peer.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InputNotifyPeer(None)

    def on_response(self, reader):
        self.peer = reader.tgread_object()

    def __repr__(self):
        return 'inputNotifyPeer#b8bc5b0c peer:InputPeer = InputNotifyPeer'

    def __str__(self):
        return '(inputNotifyPeer (ID: 0xb8bc5b0c) = (peer={}))'.format(str(self.peer))



from ....tl.mtproto_request import MTProtoRequest


class GetMessagesRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.getMessages#4222fa74 id:Vector<int> = messages.Messages"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x4222fa74

    def __init__(self, id):
        """
        :param id: Telegram type: «int». Must be a list.

        :returns messages.Messages: Instance of either Messages, MessagesSlice, ChannelMessages.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.id = id

    def on_send(self, writer):
        writer.write_int(GetMessagesRequest.constructor_id, signed=False)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.id))
        for id_item in self.id:
            writer.write_int(id_item)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetMessagesRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'messages.getMessages#4222fa74 id:Vector<int> = messages.Messages'

    def __str__(self):
        return '(messages.getMessages (ID: 0x4222fa74) = (id={}))'.format(None if not self.id else [str(_) for _ in self.id])



from ....tl.mtproto_request import MTProtoRequest


class GetDifferenceRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updates.getDifference#25939651 flags:# pts:int pts_total_limit:flags.0?int date:date qts:int = updates.Difference"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x25939651

    def __init__(self, pts, date, qts, pts_total_limit=None):
        """
        :param pts: Telegram type: «int».
        :param pts_total_limit: Telegram type: «int».
        :param date: Telegram type: «date».
        :param qts: Telegram type: «int».

        :returns updates.Difference: Instance of either DifferenceEmpty, Difference, DifferenceSlice, DifferenceTooLong.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.pts = pts
        self.pts_total_limit = pts_total_limit
        self.date = date
        self.qts = qts

    def on_send(self, writer):
        writer.write_int(GetDifferenceRequest.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.pts_total_limit else 0
        writer.write_int(flags)

        writer.write_int(self.pts)
        if self.pts_total_limit:
            writer.write_int(self.pts_total_limit)

        writer.tgwrite_date(self.date)
        writer.write_int(self.qts)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetDifferenceRequest(None, None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'updates.getDifference#25939651 flags:# pts:int pts_total_limit:flags.0?int date:date qts:int = updates.Difference'

    def __str__(self):
        return '(updates.getDifference (ID: 0x25939651) = (pts={}, pts_total_limit={}, date={}, qts={}))'.format(str(self.pts), str(self.pts_total_limit), str(self.date), str(self.qts))



from ...tl.mtproto_request import MTProtoRequest


class UpdateNewEncryptedMessage(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateNewEncryptedMessage#12bcbd9a message:EncryptedMessage qts:int = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x12bcbd9a

    def __init__(self, message, qts):
        """
        :param message: Telegram type: «EncryptedMessage».
        :param qts: Telegram type: «int».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.message = message
        self.qts = qts

    def on_send(self, writer):
        writer.write_int(UpdateNewEncryptedMessage.constructor_id, signed=False)
        self.message.on_send(writer)
        writer.write_int(self.qts)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateNewEncryptedMessage(None, None)

    def on_response(self, reader):
        self.message = reader.tgread_object()
        self.qts = reader.read_int()

    def __repr__(self):
        return 'updateNewEncryptedMessage#12bcbd9a message:EncryptedMessage qts:int = Update'

    def __str__(self):
        return '(updateNewEncryptedMessage (ID: 0x12bcbd9a) = (message={}, qts={}))'.format(str(self.message), str(self.qts))



from ...tl.mtproto_request import MTProtoRequest


class InlineBotSwitchPM(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inlineBotSwitchPM#3c20629f text:string start_param:string = InlineBotSwitchPM"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x3c20629f

    def __init__(self, text, start_param):
        """
        :param text: Telegram type: «string».
        :param start_param: Telegram type: «string».

        Constructor for InlineBotSwitchPM: Instance of InlineBotSwitchPM.
        """
        super().__init__()

        self.text = text
        self.start_param = start_param

    def on_send(self, writer):
        writer.write_int(InlineBotSwitchPM.constructor_id, signed=False)
        writer.tgwrite_string(self.text)
        writer.tgwrite_string(self.start_param)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InlineBotSwitchPM(None, None)

    def on_response(self, reader):
        self.text = reader.tgread_string()
        self.start_param = reader.tgread_string()

    def __repr__(self):
        return 'inlineBotSwitchPM#3c20629f text:string start_param:string = InlineBotSwitchPM'

    def __str__(self):
        return '(inlineBotSwitchPM (ID: 0x3c20629f) = (text={}, start_param={}))'.format(str(self.text), str(self.start_param))



from ...tl.mtproto_request import MTProtoRequest


class PaymentSavedCredentialsCard(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    paymentSavedCredentialsCard#cdc27a1f id:string title:string = PaymentSavedCredentials"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xcdc27a1f

    def __init__(self, id, title):
        """
        :param id: Telegram type: «string».
        :param title: Telegram type: «string».

        Constructor for PaymentSavedCredentials: Instance of PaymentSavedCredentialsCard.
        """
        super().__init__()

        self.id = id
        self.title = title

    def on_send(self, writer):
        writer.write_int(PaymentSavedCredentialsCard.constructor_id, signed=False)
        writer.tgwrite_string(self.id)
        writer.tgwrite_string(self.title)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PaymentSavedCredentialsCard(None, None)

    def on_response(self, reader):
        self.id = reader.tgread_string()
        self.title = reader.tgread_string()

    def __repr__(self):
        return 'paymentSavedCredentialsCard#cdc27a1f id:string title:string = PaymentSavedCredentials'

    def __str__(self):
        return '(paymentSavedCredentialsCard (ID: 0xcdc27a1f) = (id={}, title={}))'.format(str(self.id), str(self.title))



from ....tl.mtproto_request import MTProtoRequest


class SavedInfo(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    payments.savedInfo#fb8fe43c flags:# has_saved_credentials:flags.1?true saved_info:flags.0?PaymentRequestedInfo = payments.SavedInfo"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xfb8fe43c

    def __init__(self, has_saved_credentials=None, saved_info=None):
        """
        :param has_saved_credentials: Telegram type: «true».
        :param saved_info: Telegram type: «PaymentRequestedInfo».

        Constructor for payments.SavedInfo: Instance of SavedInfo.
        """
        super().__init__()

        self.has_saved_credentials = has_saved_credentials
        self.saved_info = saved_info

    def on_send(self, writer):
        writer.write_int(SavedInfo.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 1) if self.has_saved_credentials else 0
        flags |= (1 << 0) if self.saved_info else 0
        writer.write_int(flags)

        if self.saved_info:
            self.saved_info.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return SavedInfo(None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 1)) != 0:
            self.has_saved_credentials = True  # Arbitrary not-None value, no need to read since it is a flag

        if (flags & (1 << 0)) != 0:
            self.saved_info = reader.tgread_object()

    def __repr__(self):
        return 'payments.savedInfo#fb8fe43c flags:# has_saved_credentials:flags.1?true saved_info:flags.0?PaymentRequestedInfo = payments.SavedInfo'

    def __str__(self):
        return '(payments.savedInfo (ID: 0xfb8fe43c) = (has_saved_credentials={}, saved_info={}))'.format(str(self.has_saved_credentials), str(self.saved_info))



import platform
from datetime import timedelta
from hashlib import md5
from mimetypes import guess_type
from os import listdir, path
from threading import Event, RLock, Thread
from time import time, sleep
import logging

# Import some externalized utilities to work with the Telegram types and more
from . import helpers as utils
from .errors import (RPCError, InvalidDCError, FloodWaitError,
                     InvalidParameterError, ReadCancelledError)

from .network import authenticator, MtProtoSender, TcpTransport
from .parser.markdown_parser import parse_message_entities

# For sending and receiving requests
from .tl import MTProtoRequest, Session, JsonSession
from .tl.all_tlobjects import layer
from .tl.functions import (InitConnectionRequest, InvokeWithLayerRequest,
                           PingRequest)

# Required to work with different data centers
from .tl.functions.auth import (ExportAuthorizationRequest,
                                ImportAuthorizationRequest)

# Required to get the password salt
from .tl.functions.account import GetPasswordRequest

# Logging in and out
from .tl.functions.auth import (CheckPasswordRequest, LogOutRequest,
                                SendCodeRequest, SignInRequest,
                                SignUpRequest, ImportBotAuthorizationRequest)

# Initial request
from .tl.functions.help import GetConfigRequest

# Easier access to common methods
from .tl.functions.messages import (
    GetDialogsRequest, GetHistoryRequest, ReadHistoryRequest, SendMediaRequest,
    SendMessageRequest)

# For .get_me() and ensuring we're authorized
from .tl.functions.users import GetUsersRequest

# Easier access for working with media, too
from .tl.functions.upload import (
    GetFileRequest, SaveBigFilePartRequest, SaveFilePartRequest)

# All the types we need to work with
from .tl.types import (
    ChatPhotoEmpty, DocumentAttributeAudio, DocumentAttributeFilename,
    InputDocumentFileLocation, InputFile, InputFileBig, InputFileLocation,
    InputMediaUploadedDocument, InputMediaUploadedPhoto, InputPeerEmpty,
    MessageMediaContact, MessageMediaDocument, MessageMediaPhoto,
    UserProfilePhotoEmpty, InputUserSelf)

from .utils import (find_user_or_chat, get_input_peer,
                    get_appropriated_part_size, get_extension)


class TelegramClient:

    # Current TelegramClient version
    __version__ = '0.10.1'

    # region Initialization

    def __init__(self, session, api_id, api_hash, proxy=None):
        """Initializes the Telegram client with the specified API ID and Hash.

           Session can either be a `str` object (the filename for the loaded/saved .session)
           or it can be a `Session` instance (in which case list_sessions() would probably not work).
           If you don't want any file to be saved, pass `None`

           In the later case, you are free to override the `Session` class to provide different
           .save() and .load() implementations to suit your needs."""

        if api_id is None or api_hash is None:
            raise PermissionError(
                'Your API ID or Hash are invalid. Please read "Requirements" on README.rst')

        self.api_id = api_id
        self.api_hash = api_hash

        # Determine what session object we have
        # TODO JsonSession until migration is complete (by v1.0)
        if isinstance(session, str) or session is None:
            self.session = JsonSession.try_load_or_create_new(session)
        elif isinstance(session, Session):
            self.session = session
        else:
            raise ValueError(
                'The given session must either be a string or a Session instance.')

        self.transport = None
        self.proxy = proxy  # Will be used when a TcpTransport is created

        self.login_success = False

        # Safety across multiple threads (for the updates thread)
        self._lock = RLock()
        self._logger = logging.getLogger(__name__)

        # Methods to be called when an update is received
        self._update_handlers = []
        self._updates_thread_running = Event()
        self._updates_thread_receiving = Event()

        # Cache "exported" senders 'dc_id: MtProtoSender' and
        # their corresponding sessions not to recreate them all
        # the time since it's a (somewhat expensive) process.
        self._cached_senders = {}
        self._cached_sessions = {}

        # These will be set later
        self._updates_thread = None
        self.dc_options = None
        self.sender = None
        self.phone_code_hashes = {}

    # endregion

    # region Connecting

    def connect(self, reconnect=False,
                device_model=None, system_version=None,
                app_version=None, lang_code=None):
        """Connects to the Telegram servers, executing authentication if
           required. Note that authenticating to the Telegram servers is
           not the same as authenticating the desired user itself, which
           may require a call (or several) to 'sign_in' for the first time.

           Default values for the optional parameters if left as None are:
             device_model   = platform.node()
             system_version = platform.system()
             app_version    = TelegramClient.__version__
             lang_code      = 'en'
        """
        if self.transport is None:
            self.transport = TcpTransport(self.session.server_address,
                                          self.session.port, proxy=self.proxy)

        try:
            if not self.session.auth_key or (reconnect and self.sender is not None):
                self.session.auth_key, self.session.time_offset = \
                    authenticator.do_authentication(self.transport)

                self.session.save()

            self.sender = MtProtoSender(self.transport, self.session)
            self.sender.connect()

            # Set the default parameters if left unspecified
            if not device_model:
                device_model = platform.node()
            if not system_version:
                system_version = platform.system()
            if not app_version:
                app_version = self.__version__
            if not lang_code:
                lang_code = 'en'

            # Now it's time to send an InitConnectionRequest
            # This must always be invoked with the layer we'll be using
            query = InitConnectionRequest(
                api_id=self.api_id,
                device_model=device_model,
                system_version=system_version,
                app_version=app_version,
                lang_code=lang_code,
                query=GetConfigRequest())

            result = self.invoke(
                InvokeWithLayerRequest(
                    layer=layer, query=query))

            # We're only interested in the DC options,
            # although many other options are available!
            self.dc_options = result.dc_options

            self.login_success = True
            return True
        except (RPCError, ConnectionError) as error:
            # Probably errors from the previous session, ignore them
            self._logger.warning('Could not stabilise initial connection: {}'
                                 .format(error))
            return False

    def disconnect(self):
        """Disconnects from the Telegram server and stops all the spawned threads"""
        self._set_updates_thread(running=False)
        if self.sender:
            self.sender.disconnect()
            self.sender = None
        if self.transport:
            self.transport.close()
            self.transport = None

        # Also disconnect all the cached senders
        for sender in self._cached_senders.values():
            sender.disconnect()

        self._cached_senders.clear()
        self._cached_sessions.clear()

    def reconnect(self):
        """Disconnects and connects again (effectively reconnecting)"""
        self.disconnect()
        self.connect()

    # endregion

    # region Working with different Data Centers

    def _reconnect_to_dc(self, dc_id):
        """Reconnects to the specified DC ID. This is automatically
           called after an InvalidDCError is raised"""
        dc = self._get_dc(dc_id)

        self.transport.close()
        self.transport = None
        self.session.server_address = dc.ip_address
        self.session.port = dc.port
        self.session.save()

        self.connect(reconnect=True)

    def _get_dc(self, dc_id):
        """Gets the Data Center (DC) associated to 'dc_id'"""
        if not self.dc_options:
            raise ConnectionError(
                'Cannot determine the required data center IP address. '
                'Stabilise a successful initial connection first.')

        return next(dc for dc in self.dc_options if dc.id == dc_id)

    def _get_exported_sender(self, dc_id, init_connection=False):
        """Gets a cached exported MtProtoSender for the desired DC.

           If it's the first time retrieving the MtProtoSender, the
           current authorization is exported to the new DC so that
           it can be used there, and the connection is initialized.

           If after using the sender a ConnectionResetError is raised,
           this method should be called again with init_connection=True
           in order to perform the reconnection."""
        # Thanks badoualy/kotlogram on /telegram/api/DefaultTelegramClient.kt
        # for clearly showing how to export the authorization! ^^

        sender = self._cached_senders.get(dc_id)
        session = self._cached_sessions.get(dc_id)

        if sender and session:
            if init_connection:
                sender.disconnect()
                sender.connect()

            return sender
        else:
            dc = self._get_dc(dc_id)

            # Step 1. Export the current authorization to the new DC.
            export_auth = self.invoke(ExportAuthorizationRequest(dc_id))

            # Step 2. Create a transport connected to the new DC.
            #         We also create a temporary session because
            #         it's what will contain the required AuthKey
            #         for MtProtoSender to work.
            transport = TcpTransport(dc.ip_address, dc.port, proxy=self.proxy)
            session = Session(None)
            session.auth_key, session.time_offset = \
                authenticator.do_authentication(transport)

            # Step 3. After authenticating on the new DC,
            #         we can create the proper MtProtoSender.
            sender = MtProtoSender(transport, session)
            sender.connect()

            # InvokeWithLayer(InitConnection(ImportAuthorization(...)))
            init_connection = InitConnectionRequest(
                api_id=self.api_id,
                device_model=platform.node(),
                system_version=platform.system(),
                app_version=self.__version__,
                lang_code='en',
                query=ImportAuthorizationRequest(
                    export_auth.id, export_auth.bytes)
            )
            query = InvokeWithLayerRequest(layer=layer, query=init_connection)

            sender.send(query)
            sender.receive(query)

            # Step 4. We're connected and using the desired layer!
            # Don't go through this expensive process every time.
            self._cached_senders[dc_id] = sender
            self._cached_sessions[dc_id] = session

            return sender

    # endregion

    # region Telegram requests functions

    def invoke(self, request, timeout=timedelta(seconds=5), throw_invalid_dc=False):
        """Invokes a MTProtoRequest (sends and receives it) and returns its result.
           An optional timeout can be given to cancel the operation after the time delta.
           Timeout can be set to None for no timeout.

           If throw_invalid_dc is True, these errors won't be caught (useful to
           avoid infinite recursion). This should not be set to True manually."""
        if not issubclass(type(request), MTProtoRequest):
            raise ValueError('You can only invoke MtProtoRequests')

        if not self.sender:
            raise ValueError('You must be connected to invoke requests!')

        if self._updates_thread_receiving.is_set():
            self.sender.cancel_receive()

        try:
            self._lock.acquire()
            updates = []
            self.sender.send(request)
            self.sender.receive(request, timeout, updates=updates)
            for update in updates:
                for handler in self._update_handlers:
                    handler(update)

            return request.result

        except InvalidDCError as error:
            if throw_invalid_dc:
                raise

            if error.message.startswith('FILE_MIGRATE_'):
                return self.invoke_on_dc(request, error.new_dc,
                                         timeout=timeout)
            else:
                self._reconnect_to_dc(error.new_dc)
                return self.invoke(request,
                                   timeout=timeout, throw_invalid_dc=True)

        except ConnectionResetError:
            self._logger.info('Server disconnected us. Reconnecting and '
                              'resending request...')
            self.reconnect()
            self.invoke(request, timeout=timeout,
                        throw_invalid_dc=throw_invalid_dc)

        except FloodWaitError:
            self.disconnect()
            raise

        finally:
            self._lock.release()

    def invoke_on_dc(self, request, dc_id,
                     timeout=timedelta(seconds=5), reconnect=False):
        """Invokes the given request on a different DC
           by making use of the exported MtProtoSenders.

           If 'reconnect=True', then the a reconnection will be performed and
           ConnectionResetError will be raised if it occurs a second time.
        """
        try:
            sender = self._get_exported_sender(
                dc_id, init_connection=reconnect)

            sender.send(request)
            sender.receive(request)
            return request.result

        except ConnectionResetError:
            if reconnect:
                raise
            else:
                return self.invoke_on_dc(request, dc_id,
                                         timeout=timeout, reconnect=True)

    # region Authorization requests

    def is_user_authorized(self):
        """Has the user been authorized yet
           (code request sent and confirmed)?"""
        return self.session and self.get_me() is not None

    def send_code_request(self, phone_number):
        """Sends a code request to the specified phone number"""
        result = self.invoke(SendCodeRequest(phone_number, self.api_id, self.api_hash))
        self.phone_code_hashes[phone_number] = result.phone_code_hash

    def sign_in(self, phone_number=None, code=None, password=None, bot_token=None):
        """Completes the authorization of a phone number by providing the received code.

           If no phone or code is provided, then the sole password will be used. The password
           should be used after a normal authorization attempt has happened and an RPCError
           with `.password_required = True` was raised.

           To login as a bot, only `bot_token` should be provided. This should equal to the
           bot access hash provided by https://t.me/BotFather during your bot creation."""
        if phone_number and code:
            if phone_number not in self.phone_code_hashes:
                raise ValueError(
                    'Please make sure you have called send_code_request first.')

            try:
                result = self.invoke(
                    SignInRequest(phone_number, self.phone_code_hashes[
                        phone_number], code))

            except RPCError as error:
                if error.message.startswith('PHONE_CODE_'):
                    return False
                else:
                    raise
        elif password:
            salt = self.invoke(GetPasswordRequest()).current_salt
            result = self.invoke(
                CheckPasswordRequest(utils.get_password_hash(password, salt)))
        elif bot_token:
            result = self.invoke(
                ImportBotAuthorizationRequest(flags=0,
                                              api_id=self.api_id,
                                              api_hash=self.api_hash,
                                              bot_auth_token=bot_token))
        else:
            raise ValueError(
                'You must provide a phone_number and a code for the first time, '
                'and a password only if an RPCError was raised before.')

        # Ignore 'result.user', we don't need it
        self.login_success = True
        return True

    def sign_up(self, phone_number, code, first_name, last_name=''):
        """Signs up to Telegram. Make sure you sent a code request first!"""
        result = self.invoke(
            SignUpRequest(
                phone_number=phone_number,
                phone_code_hash=self.phone_code_hashes[phone_number],
                phone_code=code,
                first_name=first_name,
                last_name=last_name))

        self.session.user = result.user
        self.session.save()

    def log_out(self):
        """Logs out and deletes the current session. Returns True if everything went OK"""
        # Special flag when logging out (so the ack request confirms it)
        self.sender.logging_out = True
        try:
            self.invoke(LogOutRequest())
            self.disconnect()
            if not self.session.delete():
                return False

            self.session = None
            return True
        except (RPCError, ConnectionError):
            # Something happened when logging out, restore the state back
            self.sender.logging_out = False
            return False

    def get_me(self):
        """Gets "me" (the self user) which is currently authenticated,
           or None if the request fails (hence, not authenticated)."""
        try:
            return self.invoke(GetUsersRequest([InputUserSelf()]))[0]
        except RPCError as e:
            if e.code == 401:  # 401 UNAUTHORIZED
                return None
            else:
                raise

    @staticmethod
    def list_sessions():
        """Lists all the sessions of the users who have ever connected
           using this client and never logged out"""
        return [path.splitext(path.basename(f))[0]
                for f in listdir('.') if f.endswith('.session')]

    # endregion

    # region Dialogs ("chats") requests

    def get_dialogs(self,
                    limit=10,
                    offset_date=None,
                    offset_id=0,
                    offset_peer=InputPeerEmpty()):
        """Returns a tuple of lists ([dialogs], [entities]) with at least 'limit' items each.
           If `limit` is 0, all dialogs will be retrieved.
           The `entity` represents the user, chat or channel corresponding to that dialog"""

        r = self.invoke(
            GetDialogsRequest(
                offset_date=offset_date,
                offset_id=offset_id,
                offset_peer=offset_peer,
                limit=limit))
        return (
            r.dialogs,
            [find_user_or_chat(d.peer, r.users, r.chats) for d in r.dialogs])

    # endregion

    # region Message requests

    def send_message(self,
                     entity,
                     message,
                     markdown=False,
                     no_web_page=False):
        """Sends a message to the given entity (or input peer) and returns the sent message ID"""
        if markdown:
            msg, entities = parse_message_entities(message)
        else:
            msg, entities = message, []

        msg_id = utils.generate_random_long()
        self.invoke(
            SendMessageRequest(
                peer=get_input_peer(entity),
                message=msg,
                random_id=msg_id,
                entities=entities,
                no_webpage=no_web_page))
        return msg_id

    def get_message_history(self,
                            entity,
                            limit=20,
                            offset_date=None,
                            offset_id=0,
                            max_id=0,
                            min_id=0,
                            add_offset=0):
        """
        Gets the message history for the specified entity

        :param entity:      The entity (or input peer) from whom to retrieve the message history
        :param limit:       Number of messages to be retrieved
        :param offset_date: Offset date (messages *previous* to this date will be retrieved)
        :param offset_id:   Offset message ID (only messages *previous* to the given ID will be retrieved)
        :param max_id:      All the messages with a higher (newer) ID or equal to this will be excluded
        :param min_id:      All the messages with a lower (older) ID or equal to this will be excluded
        :param add_offset:  Additional message offset (all of the specified offsets + this offset = older messages)

        :return: A tuple containing total message count and two more lists ([messages], [senders]).
                 Note that the sender can be null if it was not found!
        """
        result = self.invoke(
            GetHistoryRequest(
                get_input_peer(entity),
                limit=limit,
                offset_date=offset_date,
                offset_id=offset_id,
                max_id=max_id,
                min_id=min_id,
                add_offset=add_offset))

        # The result may be a messages slice (not all messages were retrieved) or
        # simply a messages TLObject. In the later case, no "count" attribute is specified:
        # the total messages count is retrieved by counting all the retrieved messages
        total_messages = getattr(result, 'count', len(result.messages))

        # Iterate over all the messages and find the sender User
        users = []
        for msg in result.messages:
            for usr in result.users:
                if msg.from_id == usr.id:
                    users.append(usr)
                    break

        return total_messages, result.messages, users

    def send_read_acknowledge(self, entity, messages=None, max_id=None):
        """Sends a "read acknowledge" (i.e., notifying the given peer that we've
           read their messages, also known as the "double check ✅✅").

           Either a list of messages (or a single message) can be given,
           or the maximum message ID (until which message we want to send the read acknowledge).

           Returns an AffectedMessages TLObject"""
        if max_id is None:
            if not messages:
                raise InvalidParameterError(
                    'Either a message list or a max_id must be provided.')

            if isinstance(messages, list):
                max_id = max(msg.id for msg in messages)
            else:
                max_id = messages.id

        return self.invoke(ReadHistoryRequest(peer=get_input_peer(entity), max_id=max_id))

    # endregion

    # TODO Handle media downloading/uploading in a different session?
    #  "It is recommended that large queries (upload.getFile, upload.saveFilePart)
    #   be handled through a separate session and a separate connection"
    # region Uploading media requests

    def upload_file(self,
                    file_path,
                    part_size_kb=None,
                    file_name=None,
                    progress_callback=None):
        """Uploads the specified file_path and returns a handle which can be later used

        :param file_path: The file path of the file that will be uploaded
        :param part_size_kb: The part size when uploading the file. None = Automatic
        :param file_name: The name of the uploaded file. None = Automatic
        :param progress_callback: A callback function which takes two parameters,
                                  uploaded size (in bytes) and total file size (in bytes)
                                  This is called every time a part is uploaded
        """
        file_size = path.getsize(file_path)
        if not part_size_kb:
            part_size_kb = get_appropriated_part_size(file_size)

        if part_size_kb > 512:
            raise ValueError('The part size must be less or equal to 512KB')

        part_size = int(part_size_kb * 1024)
        if part_size % 1024 != 0:
            raise ValueError('The part size must be evenly divisible by 1024')

        # Determine whether the file is too big (over 10MB) or not
        # Telegram does make a distinction between smaller or larger files
        is_large = file_size > 10 * 1024 * 1024
        part_count = (file_size + part_size - 1) // part_size

        # Multiply the datetime timestamp by 10^6 to get the ticks
        # This is high likely going to be unique
        file_id = utils.generate_random_long()
        hash_md5 = md5()

        with open(file_path, 'rb') as file:
            for part_index in range(part_count):
                # Read the file by in chunks of size part_size
                part = file.read(part_size)

                # The SavePartRequest is different depending on whether
                # the file is too large or not (over or less than 10MB)
                if is_large:
                    request = SaveBigFilePartRequest(file_id, part_index,
                                                     part_count, part)
                else:
                    request = SaveFilePartRequest(file_id, part_index, part)

                # Invoke the file upload and increment both the part index and MD5 checksum
                result = self.invoke(request)
                if result:
                    if not is_large:
                        # No need to update the hash if it's a large file
                        hash_md5.update(part)

                    if progress_callback:
                        progress_callback(file.tell(), file_size)
                else:
                    raise ValueError('Could not upload file part #{}'.format(
                        part_index))

        # Set a default file name if None was specified
        if not file_name:
            file_name = path.basename(file_path)

        # After the file has been uploaded, we can return a handle pointing to it
        if is_large:
            return InputFileBig(
                id=file_id,
                parts=part_count,
                name=file_name)
        else:
            return InputFile(
                id=file_id,
                parts=part_count,
                name=file_name,
                md5_checksum=hash_md5.hexdigest())

    def send_photo_file(self, input_file, entity, caption=''):
        """Sends a previously uploaded input_file
           (which should be a photo) to the given entity (or input peer)"""
        self.send_media_file(
            InputMediaUploadedPhoto(input_file, caption), entity)

    def send_document_file(self, input_file, entity, caption=''):
        """Sends a previously uploaded input_file
           (which should be a document) to the given entity (or input peer)"""

        # Determine mime-type and attributes
        # Take the first element by using [0] since it returns a tuple
        mime_type = guess_type(input_file.name)[0]
        attributes = [
            DocumentAttributeFilename(input_file.name)
            # TODO If the input file is an audio, find out:
            # Performer and song title and add DocumentAttributeAudio
        ]
        # Ensure we have a mime type, any; but it cannot be None
        # «The "octet-stream" subtype is used to indicate that a body contains arbitrary binary data.»
        if not mime_type:
            mime_type = 'application/octet-stream'
        self.send_media_file(
            InputMediaUploadedDocument(
                file=input_file,
                mime_type=mime_type,
                attributes=attributes,
                caption=caption),
            entity)

    def send_media_file(self, input_media, entity):
        """Sends any input_media (contact, document, photo...) to the given entity"""
        self.invoke(
            SendMediaRequest(
                peer=get_input_peer(entity),
                media=input_media,
                random_id=utils.generate_random_long()))

    # endregion

    # region Downloading media requests

    def download_profile_photo(self,
                               profile_photo,
                               file_path,
                               add_extension=True,
                               download_big=True):
        """Downloads the profile photo for an user or a chat (including channels).
           Returns False if no photo was provided, or if it was Empty"""

        if (not profile_photo or
                isinstance(profile_photo, UserProfilePhotoEmpty) or
                isinstance(profile_photo, ChatPhotoEmpty)):
            return False

        if add_extension:
            file_path += get_extension(profile_photo)

        if download_big:
            photo_location = profile_photo.photo_big
        else:
            photo_location = profile_photo.photo_small

        # Download the media with the largest size input file location
        self.download_file_loc(
            InputFileLocation(
                volume_id=photo_location.volume_id,
                local_id=photo_location.local_id,
                secret=photo_location.secret),
            file_path)
        return True

    def download_msg_media(self,
                           message_media,
                           file_path,
                           add_extension=True,
                           progress_callback=None):
        """Downloads the given MessageMedia (Photo, Document or Contact)
           into the desired file_path, optionally finding its extension automatically
           The progress_callback should be a callback function which takes two parameters,
           uploaded size (in bytes) and total file size (in bytes).
           This will be called every time a part is downloaded"""
        if type(message_media) == MessageMediaPhoto:
            return self.download_photo(message_media, file_path, add_extension,
                                       progress_callback)

        elif type(message_media) == MessageMediaDocument:
            return self.download_document(message_media, file_path,
                                          add_extension, progress_callback)

        elif type(message_media) == MessageMediaContact:
            return self.download_contact(message_media, file_path,
                                         add_extension)

    def download_photo(self,
                       message_media_photo,
                       file_path,
                       add_extension=False,
                       progress_callback=None):
        """Downloads MessageMediaPhoto's largest size into the desired
           file_path, optionally finding its extension automatically
           The progress_callback should be a callback function which takes two parameters,
           uploaded size (in bytes) and total file size (in bytes).
           This will be called every time a part is downloaded"""

        # Determine the photo and its largest size
        photo = message_media_photo.photo
        largest_size = photo.sizes[-1]
        file_size = largest_size.size
        largest_size = largest_size.location

        if add_extension:
            file_path += get_extension(message_media_photo)

        # Download the media with the largest size input file location
        self.download_file_loc(
            InputFileLocation(
                volume_id=largest_size.volume_id,
                local_id=largest_size.local_id,
                secret=largest_size.secret),
            file_path,
            file_size=file_size,
            progress_callback=progress_callback)
        return file_path

    def download_document(self,
                          message_media_document,
                          file_path=None,
                          add_extension=True,
                          progress_callback=None):
        """Downloads the given MessageMediaDocument into the desired
           file_path, optionally finding its extension automatically.
           If no file_path is given, it will try to be guessed from the document
           The progress_callback should be a callback function which takes two parameters,
           uploaded size (in bytes) and total file size (in bytes).
           This will be called every time a part is downloaded"""
        document = message_media_document.document
        file_size = document.size

        # If no file path was given, try to guess it from the attributes
        if file_path is None:
            for attr in document.attributes:
                if type(attr) == DocumentAttributeFilename:
                    file_path = attr.file_name
                    break  # This attribute has higher preference

                elif type(attr) == DocumentAttributeAudio:
                    file_path = '{} - {}'.format(attr.performer, attr.title)

            if file_path is None:
                raise ValueError('Could not infer a file_path for the document'
                                 '. Please provide a valid file_path manually')

        if add_extension:
            file_path += get_extension(message_media_document)

        self.download_file_loc(
            InputDocumentFileLocation(
                id=document.id,
                access_hash=document.access_hash,
                version=document.version),
            file_path,
            file_size=file_size,
            progress_callback=progress_callback)
        return file_path

    @staticmethod
    def download_contact(message_media_contact, file_path, add_extension=True):
        """Downloads a media contact using the vCard 4.0 format"""

        first_name = message_media_contact.first_name
        last_name = message_media_contact.last_name
        phone_number = message_media_contact.phone_number

        # The only way we can save a contact in an understandable
        # way by phones is by using the .vCard format
        if add_extension:
            file_path += '.vcard'

        # Ensure that we'll be able to download the contact
        utils.ensure_parent_dir_exists(file_path)

        with open(file_path, 'w', encoding='utf-8') as file:
            file.write('BEGIN:VCARD\n')
            file.write('VERSION:4.0\n')
            file.write('N:{};{};;;\n'.format(first_name, last_name
                                             if last_name else ''))
            file.write('FN:{}\n'.format(' '.join((first_name, last_name))))
            file.write('TEL;TYPE=cell;VALUE=uri:tel:+{}\n'.format(
                phone_number))
            file.write('END:VCARD\n')

        return file_path

    def download_file_loc(self,
                          input_location,
                          file_path,
                          part_size_kb=64,
                          file_size=None,
                          progress_callback=None):
        """Downloads media from the given input_file_location to the specified file_path.
           If a progress_callback function is given, it will be called taking two
           arguments (downloaded bytes count and total file size)"""

        if not part_size_kb:
            if not file_size:
                raise ValueError('A part size value must be provided')
            else:
                part_size_kb = get_appropriated_part_size(file_size)

        part_size = int(part_size_kb * 1024)
        if part_size % 1024 != 0:
            raise ValueError('The part size must be evenly divisible by 1024')

        # Ensure that we'll be able to download the media
        utils.ensure_parent_dir_exists(file_path)

        # Start with an offset index of 0
        offset_index = 0
        with open(file_path, 'wb') as file:
            while True:
                # The current offset equals the offset_index multiplied by the part size
                offset = offset_index * part_size
                result = self.invoke(
                    GetFileRequest(input_location, offset, part_size))
                offset_index += 1

                # If we have received no data (0 bytes), the file is over
                # So there is nothing left to download and write
                if not result.bytes:
                    return result.type  # Return some extra information

                file.write(result.bytes)
                if progress_callback:
                    progress_callback(file.tell(), file_size)

    # endregion

    # endregion

    # region Updates handling

    def add_update_handler(self, handler):
        """Adds an update handler (a function which takes a TLObject,
          an update, as its parameter) and listens for updates"""
        if not self.sender or not self.login_success:
            raise RuntimeError("You can't add update handlers until you've "
                               "successfully logged in.")

        first_handler = not self._update_handlers
        self._update_handlers.append(handler)
        if first_handler:
            self._set_updates_thread(running=True)

    def remove_update_handler(self, handler):
        self._update_handlers.remove(handler)
        if not self._update_handlers:
            self._set_updates_thread(running=False)

    def list_update_handlers(self):
        return self._update_handlers[:]

    def _set_updates_thread(self, running):
        """Sets the updates thread status (running or not)"""
        if running == self._updates_thread_running.is_set():
            return

        # Different state, update the saved value and behave as required
        self._logger.info('Changing updates thread running status to %s', running)
        if running:
            self._updates_thread_running.set()
            if not self._updates_thread:
                self._updates_thread = Thread(
                    name='UpdatesThread', daemon=True,
                    target=self._updates_thread_method)

            self._updates_thread.start()
        else:
            self._updates_thread_running.clear()
            if self._updates_thread_receiving.is_set():
                self.sender.cancel_receive()

    def _updates_thread_method(self):
        """This method will run until specified and listen for incoming updates"""

        # Set a reasonable timeout when checking for updates
        timeout = timedelta(minutes=1)

        while self._updates_thread_running.is_set():
            # Always sleep a bit before each iteration to relax the CPU,
            # since it's possible to early 'continue' the loop to reach
            # the next iteration, but we still should to sleep.
            sleep(0.1)

            with self._lock:
                self._logger.debug('Updates thread acquired the lock')
                try:
                    self._updates_thread_receiving.set()
                    self._logger.debug('Trying to receive updates from the updates thread')
                    result = self.sender.receive_update(timeout=timeout)
                    self._logger.info('Received update from the updates thread')
                    for handler in self._update_handlers:
                        handler(result)

                except ConnectionResetError:
                    self._logger.info('Server disconnected us. Reconnecting...')
                    self.reconnect()

                except TimeoutError:
                    self._logger.debug('Receiving updates timed out')

                except ReadCancelledError:
                    self._logger.info('Receiving updates cancelled')

                except OSError:
                    self._logger.warning('OSError on updates thread, %s logging out',
                                         'was' if self.sender.logging_out else 'was not')

                    if self.sender.logging_out:
                        # This error is okay when logging out, means we got disconnected
                        # TODO Not sure why this happens because we call disconnect()…
                        self._set_updates_thread(running=False)
                    else:
                        raise

            self._logger.debug('Updates thread released the lock')
            self._updates_thread_receiving.clear()

        # Thread is over, so clean unset its variable
        self._updates_thread = None

    # endregion



from ....tl.mtproto_request import MTProtoRequest


class UpdateUsernameRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    channels.updateUsername#3514b3de channel:InputChannel username:string = Bool"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x3514b3de

    def __init__(self, channel, username):
        """
        :param channel: Telegram type: «InputChannel».
        :param username: Telegram type: «string».

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.channel = channel
        self.username = username

    def on_send(self, writer):
        writer.write_int(UpdateUsernameRequest.constructor_id, signed=False)
        self.channel.on_send(writer)
        writer.tgwrite_string(self.username)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateUsernameRequest(None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'channels.updateUsername#3514b3de channel:InputChannel username:string = Bool'

    def __str__(self):
        return '(channels.updateUsername (ID: 0x3514b3de) = (channel={}, username={}))'.format(str(self.channel), str(self.username))



from .get_config import GetConfigRequest
from .get_nearest_dc import GetNearestDcRequest
from .get_app_update import GetAppUpdateRequest
from .save_app_log import SaveAppLogRequest
from .get_invite_text import GetInviteTextRequest
from .get_support import GetSupportRequest
from .get_app_changelog import GetAppChangelogRequest
from .get_terms_of_service import GetTermsOfServiceRequest
from .set_bot_updates_status import SetBotUpdatesStatusRequest
from .get_cdn_config import GetCdnConfigRequest



from ....tl.mtproto_request import MTProtoRequest


class GetPasswordSettingsRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    account.getPasswordSettings#bc8d11bb current_password_hash:bytes = account.PasswordSettings"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xbc8d11bb

    def __init__(self, current_password_hash):
        """
        :param current_password_hash: Telegram type: «bytes».

        :returns account.PasswordSettings: Instance of PasswordSettings.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.current_password_hash = current_password_hash

    def on_send(self, writer):
        writer.write_int(GetPasswordSettingsRequest.constructor_id, signed=False)
        writer.tgwrite_bytes(self.current_password_hash)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetPasswordSettingsRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'account.getPasswordSettings#bc8d11bb current_password_hash:bytes = account.PasswordSettings'

    def __str__(self):
        return '(account.getPasswordSettings (ID: 0xbc8d11bb) = (current_password_hash={}))'.format(str(self.current_password_hash))



from ...tl.mtproto_request import MTProtoRequest


class UpdateReadHistoryOutbox(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateReadHistoryOutbox#2f2f21bf peer:Peer max_id:int pts:int pts_count:int = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x2f2f21bf

    def __init__(self, peer, max_id, pts, pts_count):
        """
        :param peer: Telegram type: «Peer».
        :param max_id: Telegram type: «int».
        :param pts: Telegram type: «int».
        :param pts_count: Telegram type: «int».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.peer = peer
        self.max_id = max_id
        self.pts = pts
        self.pts_count = pts_count

    def on_send(self, writer):
        writer.write_int(UpdateReadHistoryOutbox.constructor_id, signed=False)
        self.peer.on_send(writer)
        writer.write_int(self.max_id)
        writer.write_int(self.pts)
        writer.write_int(self.pts_count)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateReadHistoryOutbox(None, None, None, None)

    def on_response(self, reader):
        self.peer = reader.tgread_object()
        self.max_id = reader.read_int()
        self.pts = reader.read_int()
        self.pts_count = reader.read_int()

    def __repr__(self):
        return 'updateReadHistoryOutbox#2f2f21bf peer:Peer max_id:int pts:int pts_count:int = Update'

    def __str__(self):
        return '(updateReadHistoryOutbox (ID: 0x2f2f21bf) = (peer={}, max_id={}, pts={}, pts_count={}))'.format(str(self.peer), str(self.max_id), str(self.pts), str(self.pts_count))



from ...tl.mtproto_request import MTProtoRequest


class TextUrl(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    textUrl#3c2884c1 text:RichText url:string webpage_id:long = RichText"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x3c2884c1

    def __init__(self, text, url, webpage_id):
        """
        :param text: Telegram type: «RichText».
        :param url: Telegram type: «string».
        :param webpage_id: Telegram type: «long».

        Constructor for RichText: Instance of either TextEmpty, TextPlain, TextBold, TextItalic, TextUnderline, TextStrike, TextFixed, TextUrl, TextEmail, TextConcat.
        """
        super().__init__()

        self.text = text
        self.url = url
        self.webpage_id = webpage_id

    def on_send(self, writer):
        writer.write_int(TextUrl.constructor_id, signed=False)
        self.text.on_send(writer)
        writer.tgwrite_string(self.url)
        writer.write_long(self.webpage_id)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return TextUrl(None, None, None)

    def on_response(self, reader):
        self.text = reader.tgread_object()
        self.url = reader.tgread_string()
        self.webpage_id = reader.read_long()

    def __repr__(self):
        return 'textUrl#3c2884c1 text:RichText url:string webpage_id:long = RichText'

    def __str__(self):
        return '(textUrl (ID: 0x3c2884c1) = (text={}, url={}, webpage_id={}))'.format(str(self.text), str(self.url), str(self.webpage_id))



from ....tl.mtproto_request import MTProtoRequest


class SetBotPrecheckoutResultsRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.setBotPrecheckoutResults#09c2dd95 flags:# success:flags.1?true query_id:long error:flags.0?string = Bool"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x9c2dd95

    def __init__(self, query_id, success=None, error=None):
        """
        :param success: Telegram type: «true».
        :param query_id: Telegram type: «long».
        :param error: Telegram type: «string».

        :returns Bool: This type has no constructors.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.success = success
        self.query_id = query_id
        self.error = error

    def on_send(self, writer):
        writer.write_int(SetBotPrecheckoutResultsRequest.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 1) if self.success else 0
        flags |= (1 << 0) if self.error else 0
        writer.write_int(flags)

        writer.write_long(self.query_id)
        if self.error:
            writer.tgwrite_string(self.error)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return SetBotPrecheckoutResultsRequest(None, None, None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'messages.setBotPrecheckoutResults#09c2dd95 flags:# success:flags.1?true query_id:long error:flags.0?string = Bool'

    def __str__(self):
        return '(messages.setBotPrecheckoutResults (ID: 0x9c2dd95) = (success={}, query_id={}, error={}))'.format(str(self.success), str(self.query_id), str(self.error))



from ....tl.mtproto_request import MTProtoRequest


class PhoneCall(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    phone.phoneCall#ec82e140 phone_call:PhoneCall users:Vector<User> = phone.PhoneCall"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xec82e140

    def __init__(self, phone_call, users):
        """
        :param phone_call: Telegram type: «PhoneCall».
        :param users: Telegram type: «User». Must be a list.

        Constructor for phone.PhoneCall: Instance of PhoneCall.
        """
        super().__init__()

        self.phone_call = phone_call
        self.users = users

    def on_send(self, writer):
        writer.write_int(PhoneCall.constructor_id, signed=False)
        self.phone_call.on_send(writer)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for users_item in self.users:
            users_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return PhoneCall(None, None)

    def on_response(self, reader):
        self.phone_call = reader.tgread_object()
        reader.read_int()  # Vector's constructor ID
        self.users = []  # Initialize an empty list
        users_len = reader.read_int()
        for _ in range(users_len):
            users_item = reader.tgread_object()
            self.users.append(users_item)

    def __repr__(self):
        return 'phone.phoneCall#ec82e140 phone_call:PhoneCall users:Vector<User> = phone.PhoneCall'

    def __str__(self):
        return '(phone.phoneCall (ID: 0xec82e140) = (phone_call={}, users={}))'.format(str(self.phone_call), None if not self.users else [str(_) for _ in self.users])



from ...tl.mtproto_request import MTProtoRequest


class UpdateChatParticipantAdd(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updateChatParticipantAdd#ea4b0e5c chat_id:int user_id:int inviter_id:int date:date version:int = Update"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xea4b0e5c

    def __init__(self, chat_id, user_id, inviter_id, date, version):
        """
        :param chat_id: Telegram type: «int».
        :param user_id: Telegram type: «int».
        :param inviter_id: Telegram type: «int».
        :param date: Telegram type: «date».
        :param version: Telegram type: «int».

        Constructor for Update: Instance of either UpdateNewMessage, UpdateMessageID, UpdateDeleteMessages, UpdateUserTyping, UpdateChatUserTyping, UpdateChatParticipants, UpdateUserStatus, UpdateUserName, UpdateUserPhoto, UpdateContactRegistered, UpdateContactLink, UpdateNewEncryptedMessage, UpdateEncryptedChatTyping, UpdateEncryption, UpdateEncryptedMessagesRead, UpdateChatParticipantAdd, UpdateChatParticipantDelete, UpdateDcOptions, UpdateUserBlocked, UpdateNotifySettings, UpdateServiceNotification, UpdatePrivacy, UpdateUserPhone, UpdateReadHistoryInbox, UpdateReadHistoryOutbox, UpdateWebPage, UpdateReadMessagesContents, UpdateChannelTooLong, UpdateChannel, UpdateNewChannelMessage, UpdateReadChannelInbox, UpdateDeleteChannelMessages, UpdateChannelMessageViews, UpdateChatAdmins, UpdateChatParticipantAdmin, UpdateNewStickerSet, UpdateStickerSetsOrder, UpdateStickerSets, UpdateSavedGifs, UpdateBotInlineQuery, UpdateBotInlineSend, UpdateEditChannelMessage, UpdateChannelPinnedMessage, UpdateBotCallbackQuery, UpdateEditMessage, UpdateInlineBotCallbackQuery, UpdateReadChannelOutbox, UpdateDraftMessage, UpdateReadFeaturedStickers, UpdateRecentStickers, UpdateConfig, UpdatePtsChanged, UpdateChannelWebPage, UpdateDialogPinned, UpdatePinnedDialogs, UpdateBotWebhookJSON, UpdateBotWebhookJSONQuery, UpdateBotShippingQuery, UpdateBotPrecheckoutQuery, UpdatePhoneCall.
        """
        super().__init__()

        self.chat_id = chat_id
        self.user_id = user_id
        self.inviter_id = inviter_id
        self.date = date
        self.version = version

    def on_send(self, writer):
        writer.write_int(UpdateChatParticipantAdd.constructor_id, signed=False)
        writer.write_int(self.chat_id)
        writer.write_int(self.user_id)
        writer.write_int(self.inviter_id)
        writer.tgwrite_date(self.date)
        writer.write_int(self.version)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return UpdateChatParticipantAdd(None, None, None, None, None)

    def on_response(self, reader):
        self.chat_id = reader.read_int()
        self.user_id = reader.read_int()
        self.inviter_id = reader.read_int()
        self.date = reader.tgread_date()
        self.version = reader.read_int()

    def __repr__(self):
        return 'updateChatParticipantAdd#ea4b0e5c chat_id:int user_id:int inviter_id:int date:date version:int = Update'

    def __str__(self):
        return '(updateChatParticipantAdd (ID: 0xea4b0e5c) = (chat_id={}, user_id={}, inviter_id={}, date={}, version={}))'.format(str(self.chat_id), str(self.user_id), str(self.inviter_id), str(self.date), str(self.version))



from ...tl.mtproto_request import MTProtoRequest


class InputPeerChannel(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    inputPeerChannel#20adaef8 channel_id:int access_hash:long = InputPeer"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x20adaef8

    def __init__(self, channel_id, access_hash):
        """
        :param channel_id: Telegram type: «int».
        :param access_hash: Telegram type: «long».

        Constructor for InputPeer: Instance of either InputPeerEmpty, InputPeerSelf, InputPeerChat, InputPeerUser, InputPeerChannel.
        """
        super().__init__()

        self.channel_id = channel_id
        self.access_hash = access_hash

    def on_send(self, writer):
        writer.write_int(InputPeerChannel.constructor_id, signed=False)
        writer.write_int(self.channel_id)
        writer.write_long(self.access_hash)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return InputPeerChannel(None, None)

    def on_response(self, reader):
        self.channel_id = reader.read_int()
        self.access_hash = reader.read_long()

    def __repr__(self):
        return 'inputPeerChannel#20adaef8 channel_id:int access_hash:long = InputPeer'

    def __str__(self):
        return '(inputPeerChannel (ID: 0x20adaef8) = (channel_id={}, access_hash={}))'.format(str(self.channel_id), str(self.access_hash))



from ....tl.mtproto_request import MTProtoRequest


class GetFullChatRequest(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.getFullChat#3b831c66 chat_id:int = messages.ChatFull"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x3b831c66

    def __init__(self, chat_id):
        """
        :param chat_id: Telegram type: «int».

        :returns messages.ChatFull: Instance of ChatFull.
        """
        super().__init__()
        self.result = None
        self.confirmed = True  # Confirmed by default

        self.chat_id = chat_id

    def on_send(self, writer):
        writer.write_int(GetFullChatRequest.constructor_id, signed=False)
        writer.write_int(self.chat_id)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return GetFullChatRequest(None)

    def on_response(self, reader):
        self.result = reader.tgread_object()

    def __repr__(self):
        return 'messages.getFullChat#3b831c66 chat_id:int = messages.ChatFull'

    def __str__(self):
        return '(messages.getFullChat (ID: 0x3b831c66) = (chat_id={}))'.format(str(self.chat_id))



from ....tl.mtproto_request import MTProtoRequest


class MessagesSlice(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    messages.messagesSlice#0b446ae3 count:int messages:Vector<Message> chats:Vector<Chat> users:Vector<User> = messages.Messages"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0xb446ae3

    def __init__(self, count, messages, chats, users):
        """
        :param count: Telegram type: «int».
        :param messages: Telegram type: «Message». Must be a list.
        :param chats: Telegram type: «Chat». Must be a list.
        :param users: Telegram type: «User». Must be a list.

        Constructor for messages.Messages: Instance of either Messages, MessagesSlice, ChannelMessages.
        """
        super().__init__()

        self.count = count
        self.messages = messages
        self.chats = chats
        self.users = users

    def on_send(self, writer):
        writer.write_int(MessagesSlice.constructor_id, signed=False)
        writer.write_int(self.count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.messages))
        for messages_item in self.messages:
            messages_item.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for chats_item in self.chats:
            chats_item.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for users_item in self.users:
            users_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return MessagesSlice(None, None, None, None)

    def on_response(self, reader):
        self.count = reader.read_int()
        reader.read_int()  # Vector's constructor ID
        self.messages = []  # Initialize an empty list
        messages_len = reader.read_int()
        for _ in range(messages_len):
            messages_item = reader.tgread_object()
            self.messages.append(messages_item)

        reader.read_int()  # Vector's constructor ID
        self.chats = []  # Initialize an empty list
        chats_len = reader.read_int()
        for _ in range(chats_len):
            chats_item = reader.tgread_object()
            self.chats.append(chats_item)

        reader.read_int()  # Vector's constructor ID
        self.users = []  # Initialize an empty list
        users_len = reader.read_int()
        for _ in range(users_len):
            users_item = reader.tgread_object()
            self.users.append(users_item)

    def __repr__(self):
        return 'messages.messagesSlice#0b446ae3 count:int messages:Vector<Message> chats:Vector<Chat> users:Vector<User> = messages.Messages'

    def __str__(self):
        return '(messages.messagesSlice (ID: 0xb446ae3) = (count={}, messages={}, chats={}, users={}))'.format(str(self.count), None if not self.messages else [str(_) for _ in self.messages], None if not self.chats else [str(_) for _ in self.chats], None if not self.users else [str(_) for _ in self.users])



from ...tl.mtproto_request import MTProtoRequest


class KeyboardButtonUrl(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    keyboardButtonUrl#258aff05 text:string url:string = KeyboardButton"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x258aff05

    def __init__(self, text, url):
        """
        :param text: Telegram type: «string».
        :param url: Telegram type: «string».

        Constructor for KeyboardButton: Instance of either KeyboardButton, KeyboardButtonUrl, KeyboardButtonCallback, KeyboardButtonRequestPhone, KeyboardButtonRequestGeoLocation, KeyboardButtonSwitchInline, KeyboardButtonGame, KeyboardButtonBuy.
        """
        super().__init__()

        self.text = text
        self.url = url

    def on_send(self, writer):
        writer.write_int(KeyboardButtonUrl.constructor_id, signed=False)
        writer.tgwrite_string(self.text)
        writer.tgwrite_string(self.url)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return KeyboardButtonUrl(None, None)

    def on_response(self, reader):
        self.text = reader.tgread_string()
        self.url = reader.tgread_string()

    def __repr__(self):
        return 'keyboardButtonUrl#258aff05 text:string url:string = KeyboardButton'

    def __str__(self):
        return '(keyboardButtonUrl (ID: 0x258aff05) = (text={}, url={}))'.format(str(self.text), str(self.url))



from ....tl.mtproto_request import MTProtoRequest


class ChannelDifferenceTooLong(MTProtoRequest):
    """Class generated by TLObjects' generator. All changes will be ERASED. Original .tl definition below.
    updates.channelDifferenceTooLong#410dee07 flags:# final:flags.0?true pts:int timeout:flags.1?int top_message:int read_inbox_max_id:int read_outbox_max_id:int unread_count:int messages:Vector<Message> chats:Vector<Chat> users:Vector<User> = updates.ChannelDifference"""

    # Telegram's constructor ID (and unique identifier) for this class
    constructor_id = 0x410dee07

    def __init__(self, pts, top_message, read_inbox_max_id, read_outbox_max_id, unread_count, messages, chats, users, final=None, timeout=None):
        """
        :param final: Telegram type: «true».
        :param pts: Telegram type: «int».
        :param timeout: Telegram type: «int».
        :param top_message: Telegram type: «int».
        :param read_inbox_max_id: Telegram type: «int».
        :param read_outbox_max_id: Telegram type: «int».
        :param unread_count: Telegram type: «int».
        :param messages: Telegram type: «Message». Must be a list.
        :param chats: Telegram type: «Chat». Must be a list.
        :param users: Telegram type: «User». Must be a list.

        Constructor for updates.ChannelDifference: Instance of either ChannelDifferenceEmpty, ChannelDifferenceTooLong, ChannelDifference.
        """
        super().__init__()

        self.final = final
        self.pts = pts
        self.timeout = timeout
        self.top_message = top_message
        self.read_inbox_max_id = read_inbox_max_id
        self.read_outbox_max_id = read_outbox_max_id
        self.unread_count = unread_count
        self.messages = messages
        self.chats = chats
        self.users = users

    def on_send(self, writer):
        writer.write_int(ChannelDifferenceTooLong.constructor_id, signed=False)
        # Calculate the flags. This equals to those flag arguments which are NOT None
        flags = 0
        flags |= (1 << 0) if self.final else 0
        flags |= (1 << 1) if self.timeout else 0
        writer.write_int(flags)

        writer.write_int(self.pts)
        if self.timeout:
            writer.write_int(self.timeout)

        writer.write_int(self.top_message)
        writer.write_int(self.read_inbox_max_id)
        writer.write_int(self.read_outbox_max_id)
        writer.write_int(self.unread_count)
        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.messages))
        for messages_item in self.messages:
            messages_item.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.chats))
        for chats_item in self.chats:
            chats_item.on_send(writer)

        writer.write_int(0x1cb5c415, signed=False)  # Vector's constructor ID
        writer.write_int(len(self.users))
        for users_item in self.users:
            users_item.on_send(writer)

    @staticmethod
    def empty():
        """Returns an "empty" instance (all attributes are None)"""
        return ChannelDifferenceTooLong(None, None, None, None, None, None, None, None, None, None)

    def on_response(self, reader):
        flags = reader.read_int()

        if (flags & (1 << 0)) != 0:
            self.final = True  # Arbitrary not-None value, no need to read since it is a flag

        self.pts = reader.read_int()
        if (flags & (1 << 1)) != 0:
            self.timeout = reader.read_int()

        self.top_message = reader.read_int()
        self.read_inbox_max_id = reader.read_int()
        self.read_outbox_max_id = reader.read_int()
        self.unread_count = reader.read_int()
        reader.read_int()  # Vector's constructor ID
        self.messages = []  # Initialize an empty list
        messages_len = reader.read_int()
        for _ in range(messages_len):
            messages_item = reader.tgread_object()
            self.messages.append(messages_item)

        reader.read_int()  # Vector's constructor ID
        self.chats = []  # Initialize an empty list
        chats_len = reader.read_int()
        for _ in range(chats_len):
            chats_item = reader.tgread_object()
            self.chats.append(chats_item)

        reader.read_int()  # Vector's constructor ID
        self.users = []  # Initialize an empty list
        users_len = reader.read_int()
        for _ in range(users_len):
            users_item = reader.tgread_object()
            self.users.append(users_item)

    def __repr__(self):
        return 'updates.channelDifferenceTooLong#410dee07 flags:# final:flags.0?true pts:int timeout:flags.1?int top_message:int read_inbox_max_id:int read_outbox_max_id:int unread_count:int messages:Vector<Message> chats:Vector<Chat> users:Vector<User> = updates.ChannelDifference'

    def __str__(self):
        return '(updates.channelDifferenceTooLong (ID: 0x410dee07) = (final={}, pts={}, timeout={}, top_message={}, read_inbox_max_id={}, read_outbox_max_id={}, unread_count={}, messages={}, chats={}, users={}))'.format(str(self.final), str(self.pts), str(self.timeout), str(self.top_message), str(self.read_inbox_max_id), str(self.read_outbox_max_id), str(self.unread_count), None if not self.messages else [str(_) for _ in self.messages], None if not self.chats else [str(_) for _ in self.chats], None if not self.users else [str(_) for _ in self.users])
